# -*- coding: utf-8 -*-
"""emptyscript module."""

from pineboolib.qsa import formdbwidget


class interna(object):  # pylint: disable=invalid-name
    """Interna class."""

    ctx: "FormInternalObj"

    def __init__(self, context: "FormInternalObj") -> None:
        """Inicialize."""

        self.ctx = context

    def init(self) -> None:
        """Run optional inicialize script."""

        self.ctx.interna_init()


class oficial(interna):  # pylint: disable=invalid-name
    """Oficial class."""

    def __init__(self, context: "FormInternalObj") -> None:
        """Inicialize."""

        super().__init__(context)


class head(oficial):  # pylint: disable=invalid-name
    """Head class."""

    def __init__(self, context: "FormInternalObj") -> None:
        """Inicialize."""

        super().__init__(context)


class ifaceCtx(head):  # pylint: disable=invalid-name
    """IfaceCtx class."""

    def __init__(self, context: "FormInternalObj") -> None:
        """Inicialize."""

        super().__init__(context)


class FormInternalObj(formdbwidget.FormDBWidget):
    """FormInternalObj class."""

    def _class_init(self) -> None:
        """Inicialize ifaceCtx."""

        self.iface = ifaceCtx(self)

    def interna_init(self) -> None:
        """Run optional inicialize script."""
        pass



"""
Provide some functions based on data.
"""

from pineboolib.core.utils import logging
from pineboolib.application import types, qsadictmodules, load_script
from pineboolib import application
from pineboolib.application.database.orm import dummy_cursor
from pineboolib.application.database import pnsqlcursor, pnsqlquery
from pineboolib.core.utils import utils_base

import datetime

from typing import Any, Union, List, Optional, TYPE_CHECKING


if TYPE_CHECKING:
    from pineboolib.interfaces import iconnection, isqlcursor  # noqa : F401 # pragma: no cover
    from pineboolib.application import file as file_app  # noqa : F401 # pragma: no cover

LOGGER = logging.get_logger(__name__)


def next_counter(
    name_or_series: str,
    cursor_or_name: Union[str, "isqlcursor.ISqlCursor", "dummy_cursor.DummyCursor"],
    cursor_: Optional["isqlcursor.ISqlCursor"] = None,
) -> Optional[Union[str, int]]:
    """
    Return the following value of a counter type field of a table.

    This method is very useful when inserting records in which
    the reference is sequential and we don't remember which one was the last
    number used The return value is a QVariant of the field type is
    the one that looks for the last reference. The most advisable thing is that the type
    of the field be 'String' because this way it can be formatted and be
    used to generate a barcode. The function anyway
    supports both that the field is of type 'String' and of type 'double'.

    @param name Field name
    @param cursor_ Cursor to the table where the field is located.
    @return Qvariant with the following number.
    @author Andrés Otón Urbano.
    """
    """
    dpinelo: This method is an extension of nextCounter but allowing the introduction of a first
    character sequence It is useful when we want to keep different counters within the same table.
    Example, Customer Group Table: We add a prefix field, which will be a letter: A, B, C, D.
    We want customer numbering to be of type A00001, or B000023. With this function, we can
    Keep using the counter methods when we add that letter.

    This method returns the following value of a counter type field of a table for a given series.

    This method is very useful when inserting records in which
    the reference is sequential according to a sequence and we don't remember which one was the last
    number used The return value is a QVariant of the field type is
    the one that looks for the last reference. The most advisable thing is that the type
    of the field be 'String' because this way it can be formatted and be
    used to generate a barcode. The function anyway
    supports both that the field is of type 'String' and of type 'double'.

    @param series series that differentiates counters
    @param name Field name
    @param cursor_ Cursor to the table where the field is located.
    @return Qvariant with the following number.
    @author Andrés Otón Urbano.
    """

    if cursor_ is None:
        if not isinstance(cursor_or_name, (pnsqlcursor.PNSqlCursor, dummy_cursor.DummyCursor)):
            raise ValueError
        return _next_counter2(name_or_series, cursor_or_name)
    else:
        if not isinstance(cursor_or_name, str):
            raise ValueError
        return _next_counter3(name_or_series, cursor_or_name, cursor_)


def _next_counter2(
    name_: str, cursor_: Union["isqlcursor.ISqlCursor", "dummy_cursor.DummyCursor"]
) -> Optional[Union[str, int]]:
    if not cursor_:
        return None

    tmd = cursor_.metadata()

    field = tmd.field(name_)
    if field is None:
        return None

    type_ = field.type()

    if type_ not in ("string", "double"):
        return None

    _len = int(field.length())
    _where: str = cursor_.db().sqlLength(name_, _len)

    qry = pnsqlquery.PNSqlQuery(None, cursor_.db())
    qry.setForwardOnly(True)
    qry.setTablesList(tmd.name())
    qry.setSelect(name_)
    qry.setFrom(tmd.name())
    qry.setWhere(_where)
    qry.setOrderBy(name_ + " ASC")

    if not qry.exec_():
        return None

    _value: str = qry.value(0) if qry.last() else "0"
    _serie: str = "".join(digit for digit in _value if not digit.isdigit())
    _len_serie = len(_serie)
    _len_numero = _len - _len_serie
    print("***", _value, _serie, _len_serie, _len_numero)
    _numero: int = int(_value[_len_serie:]) + 1

    if type_ == "string":
        _numero_str: str = str(_numero).rjust(_len_numero, "0")
        return "%s%s" % (_serie, _numero_str)

    elif type_ == "double":
        return _numero

    return None


def _next_counter3(
    serie_: str, name_: str, cursor_: Union["isqlcursor.ISqlCursor", "dummy_cursor.DummyCursor"]
) -> Optional[Union[str, int]]:
    if not cursor_:
        return None

    tmd = cursor_.metadata()

    field = tmd.field(name_)

    if field is None:
        return None

    _type = field.type()

    _len: int = field.length() - len(serie_)
    _where: str = "length(%s)=%d AND %s" % (
        name_,
        field.length(),
        cursor_.db().connManager().manager().formatAssignValueLike(name_, "string", serie_, True),
    )

    qry = pnsqlquery.PNSqlQuery(None, cursor_.db())
    qry.setForwardOnly(True)
    qry.setTablesList(tmd.name())
    qry.setSelect(name_)
    qry.setFrom(tmd.name())
    qry.setWhere(_where)
    qry.setOrderBy(name_ + " ASC")

    if not qry.exec_():
        return None

    _numero: int = int(qry.value(0)[len(serie_) :]) if qry.last() else 0
    _numero += 1

    if _type == "string":
        _cadena: str = str(_numero)
        return _cadena.rjust(_len, "0") if len(_cadena) < _len else _cadena
    elif _type == "double":
        return _numero

    return None


def sql_select(
    from_: str,
    select_: str,
    where_: Optional[str] = None,
    table_list_: Optional[Union[str, List, types.Array]] = None,
    size_: int = 0,  # pylint: disable=unused-argument
    conn_: Union[str, "iconnection.IConnection"] = "default",
) -> Any:
    """
    Execute a query of type select, returning the results of the first record found.

    @param from_: from the query statement.
    @param select_: Select statement of the query, which will be the name of the field to return.
    @param where_: Where statement of the query.
    @param table_list_: Tableslist statement of the query. Required when more than one table is included in the from statement.
    @param size_: Number of lines found. (-1 if there is error).
    @param conn_name_ Connection name.
    @return Value resulting from the query or false if it finds nothing.
    """

    if where_ is None:
        where_ = "1 = 1"

    _qry = pnsqlquery.PNSqlQuery(None, conn_)

    if table_list_:
        _qry.setTablesList(table_list_)

    _qry.setSelect(select_)
    _qry.setFrom(from_)
    _qry.setWhere(where_)

    return False if not _qry.exec_() or not _qry.first() else _qry.value(0)


def quick_sql_select(
    from_: str,
    select_: str,
    where_: Optional[str] = "1 = 1",
    conn_: Union[str, "iconnection.IConnection"] = "default",
) -> Any:
    """
    Quick version of sqlSelect. Run the query directly without checking.Use with caution.
    """

    _qry = pnsqlquery.PNSqlQuery(None, conn_)
    return (
        _qry.value(0)
        if _qry.exec_("SELECT %s FROM %s WHERE %s " % (select_, from_, where_)) and _qry.first()
        else False
    )


def sql_insert(
    table_: str,
    field_list_: Union[str, List[str], types.Array],
    value_list_: Union[str, List, bool, int, float, types.Array],
    conn_: Union[str, "iconnection.IConnection"] = "default",
) -> bool:
    """
    Perform the insertion of a record in a table using an FLSqlCursor object.

    @param table_ Table name.
    @param field_list_ Comma separated list of field names.
    @param value_list_ Comma separated list of corresponding values.
    @param conn_name_ Connection name.
    @return True in case of successful insertion, False in any other case.
    """
    _value_list: Union[List[Any], types.Array] = (
        value_list_.split(",")
        if isinstance(value_list_, str)
        else (
            value_list_
            if isinstance(value_list_, (List, types.Array))
            else [value_list_]  # type: ignore [list-item]
        )
    )

    _field_list: Union[List[Any], types.Array] = (
        field_list_.split(",") if isinstance(field_list_, str) else field_list_
    )

    len_field_list = len(_field_list)

    if len_field_list != len(_value_list):
        return False

    _cursor = pnsqlcursor.PNSqlCursor(table_, True, conn_)
    _cursor.setModeAccess(_cursor.Insert)
    _cursor.refreshBuffer()

    for _pos in range(len_field_list):
        if _value_list[_pos] is None:
            _cursor.setNull(_field_list[_pos])
        else:
            _cursor.setValueBuffer(_field_list[_pos], _value_list[_pos])

    return _cursor.commitBuffer()


def sql_update(
    table_: str,
    field_list_: Union[str, List[str], types.Array],
    value_list_: Union[str, List, bool, int, float, types.Array],
    where_: str,
    conn_: Union[str, "iconnection.IConnection"] = "default",
) -> bool:
    """
    Modify one or more records in a table using an FLSqlCursor object.

    @param table_ Table name.
    @param field_list_ Comma separated list of field names.
    @param value_list_ Comma separated list of corresponding values.
    @param where_ Where statement to identify the records to be edited.
    @param conn_name_ Connection name.
    @return True in case of successful insertion, false in any other case.
    """

    _cursor = pnsqlcursor.PNSqlCursor(table_, True, conn_)
    _cursor.select(where_)
    _cursor.setForwardOnly(True)
    while _cursor.next():
        _cursor.setModeAccess(_cursor.Edit)
        _cursor.refreshBuffer()

        if isinstance(field_list_, (List, types.Array)):
            for _pos in range(len(field_list_)):
                _cursor.setValueBuffer(
                    field_list_[_pos],
                    (
                        value_list_[_pos]
                        if isinstance(value_list_, (List, types.Array))
                        else value_list_
                    ),
                )
        else:
            _cursor.setValueBuffer(field_list_, value_list_)

        if not _cursor.commitBuffer():
            return False

    return True


def sql_delete(
    table_: str, where_: str, conn_: Union[str, "iconnection.IConnection"] = "default"
) -> bool:
    """
    Delete one or more records in a table using an FLSqlCursor object.

    @param table_ Table name.
    @param where_ Where statement to identify the records to be deleted.
    @param conn_name_ Connection name.
    @return True in case of successful insertion, false in any other case.
    """

    _cursor = pnsqlcursor.PNSqlCursor(table_, True, conn_)

    # if not c.select(w):
    #     return False

    _cursor.select(where_)
    _cursor.setForwardOnly(True)

    while _cursor.next():
        _cursor.setModeAccess(_cursor.Del)
        _cursor.refreshBuffer()
        if not _cursor.commitBuffer():
            return False

    return True


def quick_sql_delete(
    table_: str, where_: str, conn_: Union[str, "iconnection.IConnection"] = "default"
) -> bool:
    """
    Quick version of sqlDelete. Execute the query directly without checking and without committing signals.Use with caution.
    """
    return exec_sql("DELETE FROM %s WHERE %s" % (table_, where_), conn_)


def exec_sql(sql_: str, conn_: Union[str, "iconnection.IConnection"] = "default") -> bool:
    """
    Run a query.
    """

    if application.PROJECT.conn_manager is None:
        raise Exception("Project is not connected yet")

    my_conn = application.PROJECT.conn_manager.useConn(conn_) if isinstance(conn_, str) else conn_

    try:
        last = my_conn.lastError()
        LOGGER.info("execSql: Ejecutando la consulta : %s", sql_)
        sql_ = sql_.replace("%", "%%") if my_conn.driver()._parse_porc else sql_

        my_conn.execute_query(sql_)  # noqa: F841
        return my_conn.lastError() == last
    except Exception as exc:
        LOGGER.exception("execSql: Error al ejecutar la consulta SQL: %s %s", sql_, exc)
        return False


def process_file_class(file_obj: "file_app.File") -> None:
    """Process file class."""
    file_ = open(file_obj.path(), "r", encoding="UTF-8", errors="replace")
    text_ = file_.read()
    file_.close()
    class_name = ""
    if text_.find("public_class =") > -1:
        class_name = text_[text_.find("public_class =") + 15 :].split(" ")[0][1:-1]
        if class_name.find('"') > -1:
            class_name = class_name[0 : class_name.find('"')]
        if class_name.find("'") > -1:
            class_name = class_name[0 : class_name.find("'")]

    if class_name:
        application.FILE_CLASSES[class_name] = file_obj.filename


class ClassManager(object):
    """ClassManager class."""

    def __getattr__(self, name: str) -> Any:
        """Return class."""

        class_ = None
        if name in application.FILE_CLASSES.keys():
            class_ = getattr(
                qsadictmodules.QSADictModules.qsa_dict_modules(), "%s_class" % name, None
            )
            if class_ is None:
                module_ = load_script.load_module(application.FILE_CLASSES[name])
                main_class = getattr(module_, name, None)
                qsadictmodules.QSADictModules.set_qsa_tree("%s_class" % name, main_class)
                class_ = main_class
        return class_

    def classes(self) -> List[str]:
        """Return available models list."""
        return list(application.FILE_CLASSES.keys())


def resolve_empty_qsa_value(type_: str = "double") -> Any:
    """Return empty values."""
    result: Any = None

    if type_ in ("double", "int", "uint", "serial"):
        result = 0
    elif type_ in ("string", "stringlist", "pixmap", "date", "timestamp"):
        result = ""
    elif type_ in ("unlock", "bool"):
        result = False
    elif type_ == "time":
        result = "00:00:00"
    elif type_ == "bytearray":
        result = bytearray()
    elif type_ == "json":
        result = {}

    return result


def resolve_qsa_value(type_: str, value: Any) -> Any:
    """Return formateed value."""

    result: Any = value

    if type_ in ("string", "stringlist"):
        result = value
    elif type_ == "double":
        try:
            result = float(value)
        except Exception as error:
            LOGGER.warning(str(error))

    elif type_ in ("int", "uint", "serial"):
        result = int(value)  # type: ignore [arg-type] # noqa: F821

    elif type_ == "date":
        if not isinstance(value, types.Date):
            result = types.Date(value)
    elif type_ == "time":
        if not isinstance(result, str):
            result = value.strftime("%H:%M:%S")

        if result.find(".") > -1:
            result = result[0 : result.find(".")]

    elif type_ in ("unlock", "bool"):
        if isinstance(value, str):
            result = utils_base.text2bool(value)
        else:
            result = types.boolean(value)
    elif type_ == "bytearray":
        result = bytearray(value)
    elif type_ == "timestamp":
        if isinstance(value, datetime.datetime):
            value = value.strftime("%Y-%m-%d %H:%M:%S")
        elif not isinstance(value, str):
            days, seconds = value.days, value.seconds
            hours = days * 24 + seconds // 3600
            minutes = (seconds % 3600) // 60
            seconds = seconds % 60
            value = "%s:%s:%s" % (
                hours,
                minutes if len(str(minutes)) > 1 else "0%s" % minutes,
                seconds if len(str(seconds)) > 1 else "0%s" % seconds,
            )

            value = str(value)
        if value.find(".") > -1:
            value = value[0 : value.find(".")]
        elif value.find("+") > -1:
            value = value[0 : value.find("+")]
        result = value

    else:
        try:
            result = float(value)
        except Exception:
            LOGGER.warning("Unknown type %s, value %s" % (type_, value))

    return result



"""Test_systype module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.fllegacy import systype
from pineboolib import application
from pineboolib.fllegacy.tests import fixture_path


class TestSysType(unittest.TestCase):
    """TestSysType Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_file_write(self) -> None:
        """Test FileWrite attributes."""

        from pineboolib.application import types

        qsa_sys = systype.SysType()
        txt = "avión, cañita"
        path_1 = "%s/test_systype_one_iso_8859-15.txt" % application.PROJECT.tmpdir
        path_2 = "%s/test_systype_one_utf-8.txt" % application.PROJECT.tmpdir

        qsa_sys.fileWriteIso(path_1, txt)
        qsa_sys.fileWriteUtf8(path_2, txt)

        file_1 = types.File(path_1, "ISO-8859-15")
        file_2 = types.File(path_2, "UTF-8")

        result_1 = file_1.read()
        result_2 = file_2.read()

        self.assertEqual(result_1, txt)
        self.assertEqual(result_2, txt)

    def test_translation(self) -> None:
        """Test translation function."""

        qsa_sys = systype.SysType()

        application.PROJECT.aq_app.loadTranslationFromModule("sys", "es")
        self.assertEqual(qsa_sys.translate("scripts", "hola python"), "Holaaaaa")
        self.assertEqual(qsa_sys.translate("python", "hola python sin group"), "Hola de nuevo!")

    def test_eneboopkg(self) -> None:
        """Test eneboopkgs load."""
        from pineboolib.qsa import qsa
        import os

        qsa_sys = systype.SysType()
        path = fixture_path("principal.eneboopkg")
        self.assertTrue(os.path.exists(path))
        self.assertTrue(qsa_sys.loadModules(path, False))
        qsa_sys.registerUpdate(path)

        qry = qsa.FLSqlQuery()
        qry.setTablesList("flfiles")
        qry.setSelect("count(nombre)")
        qry.setFrom("flfiles")
        qry.setWhere("1=1")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        self.assertEqual(qry.value(0), 148)

        qry_2 = qsa.FLSqlQuery()
        qry_2.setTablesList("flfiles")
        qry_2.setSelect("nombre")
        qry_2.setFrom("flfiles")
        qry_2.setWhere("nombre='impuestos.py'")
        self.assertTrue(qry_2.exec_())
        self.assertTrue(qry_2.first())
        self.assertEqual(qry_2.value(0), "impuestos.py")

    def test_run_transaction(self) -> None:
        """Test run transaction."""
        from pineboolib.qsa import qsa

        o_param_ = qsa.Object()
        fun_ = "test"
        o_param_.errorMsg = qsa.sys.translate("Error en la función %s" % fun_)
        f_1 = qsa.Function("oParam", "return true;")
        result_1 = qsa.sys.runTransaction(f_1, o_param_)
        self.assertTrue(result_1)
        f_2 = qsa.Function("oParam", 'oParam.errorMsg = "Holaa";return false;')
        result_2 = qsa.sys.runTransaction(f_2, o_param_)
        self.assertFalse(result_2)

    def test_transaction_level(self) -> None:
        from pineboolib.qsa import qsa
        from pineboolib.application.metadata import pnrelationmetadata

        sys_1 = systype.SysType()
        sys_2 = qsa.sys

        self.assertEqual(sys_1.transactionLevel(), 0)
        self.assertEqual(sys_2.transactionLevel(), 0)

        cur_areas = qsa.FLSqlCursor("flareas")

        cursor = qsa.FLSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "H")
        cursor.setValueBuffer("descripcion", "Área de prueba H")
        self.assertTrue(cursor.commitBuffer())
        rel = pnrelationmetadata.PNRelationMetaData(
            "flareas", "idarea", pnrelationmetadata.PNRelationMetaData.RELATION_1M
        )
        rel.setField("idarea")
        cur_areas.select("idarea ='H'")
        cur_areas.first()
        self.assertEqual(cur_areas.valueBuffer("idarea"), "H")
        cur_areas.setModeAccess(cur_areas.Edit)
        cur_areas.refreshBuffer()
        cur_modulos = qsa.FLSqlCursor("flmodules", True, "default", cur_areas, rel)
        cur_modulos.select()
        cur_modulos.refreshBuffer()
        self.assertEqual(sys_1.transactionLevel(), 0)
        self.assertEqual(sys_2.transactionLevel(), 0)

        cur_modulos.setModeAccess(cur_modulos.Insert)
        cur_modulos.transaction()
        self.assertEqual(sys_1.transactionLevel(), 1)
        self.assertEqual(sys_2.transactionLevel(), 1)
        self.assertTrue(cur_modulos.rollback())

    # def test_mod_main_widget(self) -> None:
    #    """Test modMainWidget."""
    #    from pineboolib.qsa import qsa

    #    mw = qsa.sys.modMainWidget("sys")
    #    self.assertTrue(mw)
    def test_translations(self) -> None:
        """Test translations functions."""

        sys = systype.SysType()

        self.assertEqual(sys.translate("MetaData", "123"), "123")

    def test_pixmap(self) -> None:
        """Text str to pixmap function."""
        from pineboolib.application.database import pnsqlcursor
        from PyQt6 import QtCore  # type: ignore[import]

        sys = systype.SysType()
        cursor = pnsqlcursor.PNSqlCursor("flmodules")
        cursor.select("1=1")
        cursor.first()
        buffer_ = cursor.buffer()
        self.assertTrue(buffer_)
        if buffer_:
            icono_txt = buffer_.value("icono")
            pixmap = sys.toPixmap(str(icono_txt))
            self.assertTrue(pixmap)
            res_txt = sys.fromPixmap(pixmap)
            self.assertTrue(res_txt.find("22 22 214 2") > -1)
            pixmap_2 = sys.scalePixmap(pixmap, 50, 50, QtCore.Qt.AspectRatioMode.KeepAspectRatio)
            self.assertTrue(pixmap_2)

    def test_project_info(self) -> None:
        """Test information functions."""
        sys = systype.SysType()

        document_ = sys.mvProjectXml()
        self.assertEqual(document_.toString(), "")

        list_modules = sys.mvProjectModules()
        self.assertEqual(list_modules, [])

        list_extensions = sys.mvProjectExtensions()
        self.assertEqual(list_extensions, [])
        self.assertEqual(sys.calculateShaGlobal(), "CD2E6BE78663D365EACDEE772DE13C6DBF0FF707")
        changes = sys.localChanges()
        self.assertEqual(changes["size"], 96)
        res_ = sys.xmlFilesDefBd()
        self.assertEqual(sys.diffXmlFilesDef(res_, res_)["size"], 0)
        self.assertTrue(res_)

    def test_project_info_2(self) -> None:
        """Test information functions."""
        from pineboolib import application

        sys = systype.SysType()
        res_1 = sys.xmlModule("flfactppal")
        self.assertTrue(res_1)
        path_ = application.PROJECT.tmpdir
        sys.exportModule("flfactppal", path_)
        sys.importModule("%s/flfactppal/flfactppal.mod" % path_)

    def test_project_others(self) -> None:
        """Test basics functions."""
        sys = systype.SysType()
        list_1 = sys.getWidgetList("formRecordclientes", "FLFieldDB")
        list_2 = sys.getWidgetList("formRecordclientes", "FLTableDB")
        list_3 = sys.getWidgetList("formRecordclientes", "Button")
        self.assertTrue(list_3)
        self.assertTrue(list_1)
        self.assertTrue(list_2)

    def test_project_dump(self) -> None:
        """Test dump class."""
        from pineboolib import application
        import sys

        ad_ = systype.AbanQDbDumper(
            application.PROJECT.conn_manager.useConn("default"), application.PROJECT.tmpdir, False
        )
        ad_.initDump()
        ad_.buildGui()
        self.assertTrue(ad_.launchProc(["dir"]))
        self.assertNotEqual(ad_.proc_.readLine().decode(sys.getdefaultencoding()), None)
        ad_.changeDirBase(".")

    def test_basic_1(self) -> None:
        """Test basic."""
        application.PROJECT.conn_manager.mainConn().setInteractiveGUI("")
        systype.SysType().selectModsDialog(["flfactppal", "flfactinfo", "flfactalma"])

    def test_exec_qsa(self) -> None:
        """Test execQSA."""

        path = fixture_path("test.qs")
        result = systype.SysType.execQSA(path)
        self.assertEqual(result, "Hola")

    def test_basic_2(self) -> None:
        """test basic2."""

        self.assertFalse(systype.SysType.modMainWidget("sys"))
        self.assertEqual(systype.SysType().questionMsgBox("msg"), True)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



# -*- coding: utf-8 -*-
"""
AQBoolFlagState package.

Save cursor states.
"""

from pineboolib.application.acls import pnboolflagstate


class AQBoolFlagState(pnboolflagstate.PNBoolFlagState):
    """AQBollFlagState Class."""

    pass


class AQBoolFlagStateList(pnboolflagstate.PNBoolFlagStateList):
    """AQBoolFlagStateList Class."""

    pass



"""Main setup script."""

import setuptools  # type: ignore
import pathlib
import subprocess
from pineboolib import application


with open("requirements.txt") as f:
    required = f.read().splitlines()

prj_ = application.PROJECT
prj_.load_version()
version_ = application.PINEBOO_VER
# Create/Update translations

languages = ["es", "en", "ca", "de", "fr", "gl", "it", "pt"]
lang_path = pathlib.Path("pineboolib")
files = [str(fil) for fil in lang_path.glob("**/*.py")]
exclude_uis = [
    "pineboolib/application/packager/tests/fixtures/principal/forms/agentes.ui",
    "pineboolib/application/packager/tests/fixtures/principal/forms/flfactppal.ui",
    "pineboolib/application/parsers/parser_ui/tests/fixtures/main_form_qt3.ui",
    "pineboolib/application/parsers/parser_ui/tests/fixtures/form_record_qt3.ui",
    "pineboolib/fllegacy/forms/FLWidgetReportViewer.ui",
]
files.extend([str(fil) for fil in lang_path.glob("**/*.ui")])
for exclude in exclude_uis:
    files.remove(exclude)

for lang in languages:
    ts_file = pathlib.Path("pineboolib/system_module/translations/sys.%s.ts" % lang).absolute()
    if subprocess.call(["pylupdate6", "-ts", ts_file, *files]):
        raise Exception("Error updating %s file!" % ts_file)


with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pineboo",
    version=version_,
    author="David Martínez Martí, José A. Fernández Fernández",
    author_email="deavidsedice@gmail.com, aullasistemas@gmail.com",
    description="ERP replacement for Eneboo written in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deavid/pineboo",
    packages=setuptools.find_packages(),
    package_data={
        "pineboolib": ["py.typed"],
        "pineboolib.application.staticloader.ui": ["*.ui"],
        "pineboolib.core.fonts.Noto_Sans": ["*"],
        "pineboolib.core.images.icono_pi": ["*"],
        "pineboolib.core.images.icons": ["*"],
        "pineboolib.core.images.splashscreen": ["*"],
        "pineboolib.system_module": ["*"],
        "pineboolib.system_module.forms": ["*.ui"],
        "pineboolib.system_module.queries": ["*.qry"],
        "pineboolib.system_module.tables": ["*.mtd"],
        "pineboolib.system_module.translations": ["*.ts"],
        "pineboolib.loader.dlgconnect": ["*.ui"],
        "pineboolib.plugins.dgi.dgi_qt.dgi_objects.dlg_about": ["*.ui"],
        "pineboolib.plugins.mainform.eneboo": ["*.ui"],
        "pineboolib.plugins.mainform.eneboo_mdi": ["*.ui"],
        "pineboolib.fllegacy.forms": ["*.ui"],
    },
    install_requires=required,
    keywords="erp pineboo eneboo accounting sales warehouse",
    python_requires=">=3.8",
    entry_points={
        "console_scripts": [
            "pineboo-parse=pineboolib.application.parsers.parser_qsa.postparse:main",
            "pineboo-pyconvert=pineboolib.application.parsers.parser_qsa.pyconvert:main",
            "pineboo-core=pineboolib.loader.main:startup_no_x",
            "pineboo-packager=pineboolib.application.packager.pnpackager:main",
            "pineboo-daemon=pineboolib.application.utils.service:main",
            "pineboo=pineboolib.loader.main:startup",
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Environment :: X11 Applications :: Qt",
        "Topic :: Office/Business :: Financial :: Accounting",
        "Typing :: Typed",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Natural Language :: Spanish",
        "Operating System :: OS Independent",
    ],
)



# -*- coding: utf-8 -*-
"""PNRelationMetaData manages relations between tables."""

from typing import Union


class PNRelationMetaData:
    """PNRelationMetaData Class."""

    """
    Constantes de tipos de cardinalidades de una relacion
    """

    RELATION_1M = "1M"
    RELATION_M1 = "M1"

    count_ = 0

    private: "PNRelationMetaDataPrivate"

    def __init__(
        self,
        other_or_foreign_table: Union["PNRelationMetaData", str],
        foreign_field: str = "",
        cardinality: str = "",
        delete_cascade: bool = False,
        update_cascade: bool = False,
        check_integrity: bool = True,
    ) -> None:
        """
        Initialize the relation.

        @param other_or_relation.  Related foreign table or other PNRelationMetaData.
        @param foreign_field Related foreign field.
        @param cardinality Cardinality of the relation.
        @param delete_cascade Deleted in cascade, only taken into account in M1 cardinalities. Defaul False
        @param update_cascade Cascade updates, only taken into account in M1 cardinalities. Default False
        @param check_integrity Integrity checks on the relation. Default True
        """
        if isinstance(other_or_foreign_table, str):
            foreign_table: str = other_or_foreign_table
            self.inicializeNewFLRelationMetaData(
                foreign_table,
                foreign_field,
                cardinality,
                delete_cascade,
                update_cascade,
                check_integrity,
            )
        else:
            self.inicializeFromFLRelationMetaData(other_or_foreign_table)

        ++self.count_

    def inicializeNewFLRelationMetaData(
        self,
        foreign_table: str,
        foreign_field: str,
        relation_cardinality: str,
        delete_cascade: bool,
        update_cascade: bool,
        check_integrity: bool,
    ) -> None:
        """
        Fill in the relation data.

        @param foreign_table Related foreign table.
        @param foreign_field Related foreign field.
        @param relation_cardinality Cardinality of the relation.
        @param delete_cascade Deleted in cascade, only taken into account in M1 cardinalities.
        @param update_cascade Cascade updates, only taken into account in M1 cardinalities.
        @param check_integrity Integrity checks on the relation.
        """

        self.private = PNRelationMetaDataPrivate(
            foreign_table,
            foreign_field,
            relation_cardinality,
            delete_cascade,
            update_cascade,
            check_integrity,
        )

    def inicializeFromFLRelationMetaData(self, other: "PNRelationMetaData"):
        """
        Fill in the data from another relation.

        @param other. original PNRelationMetaData.
        """

        self.private = PNRelationMetaDataPrivate()
        self.copy(other)

    def setField(self, field_name: str) -> None:
        """
        Set the name of the related field.

        @param fN Related field name.
        """

        self.private.field_ = field_name.lower()

    def field(self) -> str:
        """
        Get in the name of the related field.

        @return Returns the name of the related field
        """

        return self.private.field_

    def foreignTable(self) -> str:
        """
        Get the name of the foreign table.

        @return Returns the name of the database table with which it is related
        """

        return self.private._foreign_table

    def foreignField(self) -> str:
        """
        Get the name of the foreign field.

        @return Returns the name of the foreign table field with which it is related
        """

        return self.private._foreign_field

    def cardinality(self) -> str:
        """
        Get the cardinality of the relationship.

        @return Returns the cardinality of the relationship, looking from the table where define this object towards the outside
        """

        return self.private._cardinality

    def deleteCascade(self) -> bool:
        """
        Get if the relationship implies cascaded deletions, it is only taken into account in M1 cardinalities.

        @return Returns TRUE if the relationship implies cascaded deletions, FALSE otherwise
        """

        return self.private._delete_cascade and self.private._cardinality == self.RELATION_M1

    def updateCascade(self) -> bool:
        """
        Get if the relationship involves cascade modifications, it is only taken into account in M1 cardinalities.

        @return Returns TRUE if the relationship implies cascading modifications, FALSE otherwise
        """

        return self.private._update_cascade and self.private._cardinality == self.RELATION_M1

    def checkIn(self) -> bool:
        """
        Get if the integrity rules on the relationship should be applied.
        """

        return self.private._check_integrity

    def copy(self, other: "PNRelationMetaData") -> None:
        """Copy a PNRelationMetaData to another."""

        if other is not self and isinstance(other, PNRelationMetaData):
            self.private.field_ = other.private.field_
            self.private._foreign_table = other.private._foreign_table
            self.private._foreign_field = other.private._foreign_field
            self.private._cardinality = other.private._cardinality
            self.private._delete_cascade = other.private._delete_cascade
            self.private._update_cascade = other.private._update_cascade
            self.private._check_integrity = other.private._check_integrity


class PNRelationMetaDataPrivate:
    """PNRelationMetaDataPrivate Class."""

    """
    Nombre del campo a relacionar
    """

    field_: str

    """
    Nombre de la tabla foránea a relacionar
    """
    _foreign_table: str

    """
    Nombre del campo foráneo relacionado
    """
    _foreign_field: str

    """
    Cardinalidad de la relación
    """
    _cardinality: str

    """
    Indica si los borrados serán en cascada, en relaciones M1
    """
    _delete_cascade: bool

    """
    Indica si las modificaciones serán en cascada, en relaciones M1
    """
    _update_cascade: bool

    """
    Indica si se deben aplicar la reglas de integridad en esta relación
    """
    _check_integrity: bool

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the class."""

        self.field_ = ""

        if len(args) == 0:
            self.inicializeFLRelationMetaDataPrivate()
        else:
            self.inicializeNewFLRelationMetaDataPrivate(*args)

    def inicializeNewFLRelationMetaDataPrivate(
        self,
        foreign_table: str,
        foreign_field: str,
        relation_cardinality: str,
        delete_cascade: bool,
        update_cascade: bool,
        check_integrity: bool,
    ) -> None:
        """Fill initial values ​​with given values."""

        self._foreign_table = foreign_table.lower()
        self._foreign_field = foreign_field.lower()
        self._cardinality = relation_cardinality
        self._delete_cascade = delete_cascade
        self._update_cascade = update_cascade
        self._check_integrity = check_integrity

    def inicializeFLRelationMetaDataPrivate(self) -> None:
        """Initialize the empty class."""

        pass



"""
Proxy Module.
"""
from pineboolib import logging
from pineboolib.core import garbage_collector

from typing import Any, Optional, Dict, Callable, TYPE_CHECKING
import threading

if TYPE_CHECKING:
    from pineboolib.fllegacy.flformdb import FLFormDB  # noqa: F401 # pragma: no cover
    from pineboolib.qsa import formdbwidget  # pragma: no cover

LOGGER = logging.get_logger(__name__)


class DelayedObjectProxyLoader(object):
    """
    Delay load of an object until its first accessed.

    This is used to create entities such "formclientes" or "flfactppal" ahead of time and
    publish them in pineboolib.qsa.qsa so the code can freely call flfactppal.iface.XXX.

    Once the first attribute is called, the object is loaded.

    QSA Code should avoid calling directly "formclientes" and instead use QSADictModules or SafeQSA
    """

    def __init__(
        self,
        obj: Callable[..., "formdbwidget.FormDBWidget"],
        name: Optional[str] = None,
        *args: str,
        **kwargs: str
    ) -> None:
        """Initialize."""
        LOGGER.trace("obj: %r", obj)
        self._name: str = name or "unnamed-loader"
        self._obj = obj
        self._args = args
        self._kwargs = kwargs
        self.loaded_obj: Dict[int, Optional["formdbwidget.FormDBWidget"]] = {}

    def __load(self) -> "formdbwidget.FormDBWidget":
        """
        Load a new object.

        @return objeto nuevo o si ya existe , cacheado
        """
        list_name = self._name.split(".")
        id_thread = threading.current_thread().ident

        # limpieza de objetos
        list_threads = list([thread.ident for thread in threading.enumerate()])
        for key in list(self.loaded_obj.keys()):
            if key not in list_threads:
                obj_ = self.loaded_obj[key]
                self.loaded_obj[key] = None
                del self.loaded_obj[key]
                garbage_collector.check_delete(obj_, "proxy.widget")

        if not list_name[-1].startswith("formRecord"):
            if id_thread in self.loaded_obj.keys():
                if getattr(
                    self.loaded_obj[id_thread], "_loader", True  # type: ignore [index] # noqa: F821
                ):
                    return self.loaded_obj[  # type: ignore [index, return-value] # noqa: F821, F723
                        id_thread  # type: ignore [index, return-value] # noqa: F821, F723
                    ]

        self.loaded_obj[id_thread] = self._obj(  # type: ignore [index] # noqa: F821
            *self._args, **self._kwargs
        )

        LOGGER.debug(
            "DelayedObjectProxyLoader: name: %s, object: %s( *%s **%s) ---> %s",
            self._name,
            self._obj,
            self._args,
            self._kwargs,
            self.loaded_obj[id_thread],  # type: ignore [index] # noqa: F821
        )

        return_object = self.loaded_obj[id_thread]  # type: ignore [index] # noqa: F821

        if return_object is None:
            del self.loaded_obj[id_thread]  # type: ignore [arg-type] # noqa: F821
            raise Exception("Failed to load object")
        else:
            return_object.set_proxy_parent(self)
            return return_object

    def class_(self):
        """Return class."""
        return self._obj(*self._args, **self._kwargs)

    def __getattr__(self, name: str) -> Any:  # Solo se lanza si no existe la propiedad.
        """
        Return attribute or method from internal object.

        @param name. Nombre del la función buscada
        @return el objecto del XMLAction afectado
        """

        obj_ = self.__load()
        return getattr(obj_, name, getattr(obj_, name, None))



"""Aqobjectquerylist module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing


class TestAQObjectQueryList(unittest.TestCase):
    """TestAQObjectQueryList Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_aqobject_query_list(self) -> None:
        """Test AQObjectQueryList function."""
        from pineboolib import application
        from pineboolib.qsa import qsa
        from pineboolib.plugins.mainform.eneboo import eneboo

        main_form_class = getattr(eneboo, "MainForm", None)
        self.assertTrue(main_form_class)
        application.PROJECT.main_window = main_form_class()  # type: ignore[misc]
        self.assertTrue(application.PROJECT.main_window)
        if application.PROJECT.main_window is not None:
            application.PROJECT.main_window.initScript()
            application.PROJECT.main_window.show()

            list_ = qsa.AQObjectQueryList(
                application.PROJECT.main_window, "QAction", None, False, True
            )
            self.assertTrue(len(list_) > 65, "El tamaño devuelto es %s" % len(list_))

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Test reinit module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa
from pineboolib.plugins.mainform import eneboo
from pineboolib import application


class TestFLReinit(unittest.TestCase):
    """TestFLReinit class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""

        init_testing()

    def test_basic1(self) -> None:
        """Test reinit function."""

        application.PROJECT.main_window = eneboo.MainForm()
        application.PROJECT.main_window.initScript()

        reinit_func = qsa.from_project("formflreinit").main
        self.assertTrue(reinit_func)
        reinit_func()

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



"""FLDateEdit module."""

from PyQt6 import QtWidgets, QtCore  # type: ignore[import]

import unittest

from pineboolib.fllegacy import fldateedit
from pineboolib.loader.main import init_testing, finish_testing


class TestFLDateEdit(unittest.TestCase):
    """Test FLDataTable class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self):
        """Test basic 1."""
        parent = QtWidgets.QWidget()

        control = fldateedit.FLDateEdit(parent, "nombre")
        self.assertTrue(control is not None)
        date = QtCore.QDate.fromString(str("01-01-2000"), control.DMY)
        self.assertTrue(date is not None)
        control.setDate(date)
        control.setOrder("dd.MM.yyyy")
        # self.assertTrue(len(str(control.getDate()).split(".")) == 3)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""
Module for Manager class.
"""
# from PyQt6 import QtWidgets
from typing import Any, Iterable, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces import dgi_schema  # pragma: no cover


class Manager(object):
    """
    Creates a sensible DGI interface connection to show messages.

    Mainly used to display progress in splash screen
    """

    _dgi: Optional["dgi_schema.dgi_schema"] = None

    def __init__(self, dgi: "dgi_schema.dgi_schema") -> None:
        """Create a Manager with specified DGI."""
        self._dgi = dgi

    def send(
        self, type_: str, function_: Optional[str] = None, data_: Optional[Iterable[Any]] = None
    ) -> Any:
        """Send a progress event to the manager."""

        ret_ = None

        if self._dgi is not None:
            obj_ = getattr(self._dgi, type_, None)

            if obj_ is not None:
                attr_ = getattr(obj_, function_, None) if function_ is not None else obj_

                if attr_ is not None:
                    if not data_:
                        ret_ = attr_()
                    else:
                        ret_ = attr_(*data_)

        return ret_



"""Test_flutil module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing


class TestFLUtil(unittest.TestCase):
    """Test FLUtil class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_translate(self) -> None:
        """Test Translations."""

        from pineboolib.qsa import qsa

        qsa.aqApp.loadTranslationFromModule("sys", "es")

        util = qsa.FLUtil()
        self.assertEqual(util.translate("scripts", "single"), "simple!")
        self.assertEqual(util.translate("scripts", "variable %s") % "uno", "con variable uno!")
        self.assertEqual(
            util.translate("scripts", "multiple variable %s , %s") % ("uno", "dos"),
            "Test uno y dos",
        )

    def test_sqlSelect(self) -> None:
        """Test sqlSelect."""

        from pineboolib.qsa import qsa

        cursor_areas = qsa.FLSqlCursor("flareas")
        cursor_areas.setModeAccess(cursor_areas.Insert)
        cursor_areas.refreshBuffer()
        cursor_areas.setValueBuffer("idarea", "Y")
        cursor_areas.setValueBuffer("descripcion", "123oX")
        self.assertTrue(cursor_areas.commitBuffer())
        cx_ = cursor_areas.db()

        util = qsa.FLUtil()
        self.assertEqual(util.sqlSelect("flareas", "descripcion", "idarea='Y'"), "123oX")
        self.assertEqual(
            util.sqlSelect("flareas", "descripcion", "idarea='Y'", ["flareas"]), "123oX"
        )
        self.assertEqual(
            util.sqlSelect("flareas", "descripcion", "idarea='Y'", ["flareas"], 0), "123oX"
        )
        self.assertEqual(
            util.sqlSelect("flareas", "descripcion", "idarea='Y'", ["flareas"], cx_), "123oX"
        )
        self.assertEqual(
            util.sqlSelect("flareas", "descripcion", "idarea='Y'", ["flareas"], 0, "default"),
            "123oX",
        )

        self.assertEqual(
            util.sqlSelect("flareas", "descripcion", "idarea='Y'", ["flareas"], 0, cx_), "123oX"
        )

    def test_basic_1(self) -> None:
        """ "Test basic 1."""

        from pineboolib.qsa import qsa

        util = qsa.FLUtil()

        self.assertEqual(util.partDecimal(1.21), 21)
        self.assertEqual(util.partInteger(12.345), 12)
        result = "CIENTO VEINTITRES MILLONES CUATROCIENTOS CINCUENTA Y SEIS MIL SETECIENTOS OCHENTA"
        result += " Y NUEVE EUROS CON VEINTIUN CÉNTIMOS"
        self.assertEqual(util.enLetraMonedaEuro(123456789.21), result)
        self.assertEqual(util.enLetraMonedaEuro(100), "CIEN EUROS")
        self.assertEqual(util.enLetraMonedaEuro(5000), "CINCO MIL EUROS")
        self.assertEqual(util.enLetraMonedaEuro(6002), "SEIS MIL DOS EUROS")
        self.assertEqual(util.enLetraMonedaEuro(13), "TRECE EUROS")
        self.assertEqual(util.enLetraMonedaEuro(-10), "MENOS DIEZ EUROS")
        self.assertEqual(util.enLetraMonedaEuro(5125), "CINCO MIL CIENTO VEINTICINCO EUROS")
        self.assertEqual(util.letraDni(12345678), "Z")
        self.assertEqual(util.nombreCampos("flareas"), ["3", "bloqueo", "idarea", "descripcion"])
        self.assertEqual(util.calcularDC(30660001), "2")
        self.assertEqual(util.formatoMiles("12345"), "12.345")
        self.assertEqual(util.formatoMiles("-12.05"), "-12,05")
        self.assertFalse(util.numCreditCard("5539110012141618"))
        self.assertTrue(util.numCreditCard("3716820019271998"))

    def test_dates(self) -> None:
        """Test dates functions."""
        from pineboolib.qsa import qsa

        util = qsa.FLUtil()

        self.assertEqual(str(util.addDays("30-11-2020", 2)), "2020-12-02T00:00:00")
        self.assertEqual(str(util.addDays("2021-12-31", 2)), "2022-01-02T00:00:00")
        self.assertEqual(str(util.addDays("2021-12-31T00:01:00", 2)), "2022-01-02T00:00:00")
        self.assertEqual(str(util.addMonths("2021-12-31", 12)), "2022-12-31T00:00:00")
        self.assertEqual(str(util.addMonths("01-10-2021", 12)), "2022-10-01T00:00:00")
        self.assertEqual(str(util.addYears("01-10-2021", 12)), "2033-10-01T00:00:00")
        self.assertEqual(str(util.addYears("2021-10-01", 12)), "2033-10-01T00:00:00")
        self.assertEqual(util.daysTo(qsa.Date("2019-07-12"), qsa.Date("2019-07-15")), 3)
        self.assertEqual(util.daysTo(qsa.Date("2019-07-01"), qsa.Date("2019-11-15")), 137)
        self.assertEqual(util.daysTo(qsa.Date("2019-07-01"), qsa.Date("2019-03-15")), -108)
        self.assertEqual(util.dateAMDtoDMA("2019-12-01"), "01-12-2019")
        self.assertEqual(util.dateAMDtoDMA("20"), "")
        self.assertEqual(util.dateDMAtoAMD("20"), "")

    def test_basic_2(self) -> None:
        """Test basic 2."""
        from PyQt6 import QtCore  # type: ignore[import]

        from pineboolib.fllegacy import flutil

        util = flutil.FLUtil()

        self.assertEqual(util.buildNumber("123.4533", "", 2), "123.45")
        self.assertEqual(util.buildNumber("123.451000", "", 3), "123.451")
        self.assertEqual(util.buildNumber("123.451000", "", 1), "123.5")
        self.assertEqual(util.buildNumber("123.451000", "", 2), "123.45")
        self.assertEqual(util.buildNumber("-123.451000", "", 2), "-123.45")
        self.assertEqual(util.buildNumber("-123.4", "", 2), "-123.40")

        self.assertEqual(util.nameBD(), "temp_db")
        self.assertEqual(util.nameUser(), "memory_user")
        self.assertEqual(util.getIdioma(), QtCore.QLocale().name()[:2])

    def test_basic_3(self) -> None:
        """Test basic 3."""
        from pineboolib.fllegacy import flutil

        util = flutil.FLUtil()
        self.assertEqual(util.roundFieldValue("100.0193", "fltest", "double_field"), "100.02")
        self.assertEqual(util.sha1("123456"), "7C4A8D09CA3762AF61E59520943DC26494F8941B")

    def test_sql(self) -> None:
        """Test sql functions."""
        from pineboolib.fllegacy import flutil, flsqlcursor
        from pineboolib.qsa import qsa

        ar_1 = qsa.Array("idarea", "descripcion")
        ar_2 = qsa.Array("G", "area test 2")

        util = flutil.FLUtil()
        self.assertTrue(util.sqlInsert("flareas", ["idarea", "descripcion"], ["T", "area test 1"]))
        self.assertTrue(util.sqlInsert("flareas", ar_1, ar_2))
        cur = flsqlcursor.FLSqlCursor("flareas")
        cur.select()
        self.assertEqual(cur.size(), 2)

        res_1 = util.sqlSelect("flareas", "COUNT(idarea)", "1=1")
        self.assertEqual(res_1, 2)

        res_2 = util.quickSqlSelect("flareas", "COUNT(idarea)", "1=1")
        self.assertEqual(res_2, 2)

        self.assertTrue(
            util.sqlUpdate("flareas", "descripcion", "area test modificado", "idarea ='T'")
        )
        self.assertEqual(
            util.quickSqlSelect("flareas", "descripcion", "idarea ='T'"), "area test modificado"
        )

        self.assertTrue(util.execSql("SELECT COUNT(*) FROM flareas WHERE 1=1", "default"))
        self.assertFalse(util.execSql("SELECT COUNT(*) FROG flareas WHERE 1=1", "default"))

        self.assertTrue(util.sqlDelete("flareas", "idarea ='T'", "default"))
        cur.refresh()
        self.assertEqual(cur.size(), 1)
        self.assertTrue(util.quickSqlDelete("flareas", "idarea ='G'", "default"))
        cur.refresh()
        self.assertEqual(cur.size(), 0)

    def test_field_functions(self) -> None:
        """Test field functions."""
        from pineboolib.fllegacy import flutil

        util = flutil.FLUtil()

        self.assertEqual(util.formatValue("string", "uno", True), "'UNO'")
        self.assertEqual(util.formatValue("uint", 1233, False), "1233")
        self.assertEqual(util.formatValue("double", 1233, False), "1233")
        self.assertEqual(util.formatValue("bool", True, False), "1")

        self.assertTrue(util.fieldDefaultValue("bloqueo", "flareas"))
        self.assertFalse(util.fieldIsCompoundKey("idarea", "flareas"))
        self.assertTrue(util.fieldIsPrimaryKey("idarea", "flareas"))
        self.assertTrue(util.fieldAllowNull("icono", "flmodules"))
        self.assertFalse(util.fieldAllowNull("idarea", "flareas"))
        self.assertEqual(util.fieldAliasToName("Icono", "flmodules"), "icono")
        self.assertEqual(util.fieldNameToAlias("icono", "flmodules"), "Icono")
        self.assertEqual(util.fieldType("icono", "flmodules"), 6)
        self.assertEqual(util.fieldLength("descripcion", "flareas"), 100)
        self.assertEqual(util.fieldLength("idarea", "flareas"), 15)
        self.assertEqual(util.fieldLength("bloqueo", "flareas"), 0)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Test_load_script module."""
import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application import load_script
from pineboolib.application.tests import fixture_path

import os


class TestLoadScript(unittest.TestCase):
    """TestVirtualExists Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic1(self) -> None:
        """Test basic 1."""

        db_path = fixture_path(".")
        static_path = fixture_path(".")

        file_name = "test_file.txt"

        with open(os.path.join(db_path, file_name), "w"):
            pass

        file_path = fixture_path(file_name)
        load_script._build_static_flag(file_path, db_path, static_path)

        f1_ = open(file_path, "rb")
        data = f1_.read()
        f1_.close()
        self.assertNotEqual(data.decode(), "")

    def test_basic2(self) -> None:
        """Test basic 2."""

        file_name = "test_file.txt"
        file_path = fixture_path(file_name)
        load_script._resolve_flag(file_path, "uno", "dos")

        load_script._remove(file_path)
        self.assertFalse(os.path.exists(file_path))

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Flserial por module."""

from PyQt6 import QtCore, QtSerialPort  # type: ignore[import]
from pineboolib.core import decorators


class BaudRateType(object):
    """BaudRateType class."""

    BAUD50: int = -1
    BAUD75: int = -1
    BAUD110: int = -1
    BAUD134: int = -1
    BAUD150: int = -1
    BAUD200: int = -1
    BAUD300: int = -1
    BAUD600: int = -1
    BAUD1200: int = 1200
    BAUD1800: int = -1
    BAUD2400: int = 2400
    BAUD4800: int = 4800
    BAUD9600: int = 9600
    BAUD14400: int = -1
    BAUD19200: int = 19200
    BAUD38400: int = 38400
    BAUD56000: int = -1
    BAUD57600: int = 57600
    BAUD76800: int = -1
    BAUD115200: int = 115200
    BAUD128000: int = -1
    BAUD256000: int = -1


class DataBitsType(object):
    """DataBitsType class."""

    DATA_5: int = 5
    DATA_6: int = 6
    DATA_7: int = 7
    DATA_8: int = 8


class ParityType(object):
    """ParityType class."""

    PAR_NONE: int = 0
    PAR_EVEN: int = 2
    PAR_ODD: int = 3
    PAR_SPACE: int = 4
    PAR_MARK: int = 5


class StopBitType(object):
    """StopBitType class."""

    STOP_1: int = 1
    STOP_1_5: int = 3
    STOP_2: int = 2


class FlowType(object):
    """FlowType class."""

    FLOW_OFF: int = 0
    FLOW_HARDWARE: int = 1
    FLOW_XONXOFF: int = 2


class FLSerialPort(QtCore.QObject, BaudRateType, DataBitsType, ParityType, StopBitType, FlowType):
    """FLSerialPort class."""

    _obj: "QtSerialPort.QSerialPort"

    def __init__(self, port_name: str) -> None:
        """Inicialize."""
        super().__init__()
        if str(QtCore.QSysInfo()) == "ios":
            from pineboolib.q3widgets.messagebox import MessageBox

            MessageBox.information(
                None,
                self.tr("Opción deshabilitada"),
                self.tr("FLSerialPort no está disponible para IOS"),
                MessageBox.Ok,
            )
            return
        else:
            from PyQt6 import QtSerialPort  # type: ignore [attr-defined] # noqa: F821

            self._obj = QtSerialPort.QSerialPort(port_name)

    def setBaudRate(self, baud_rate: int) -> None:
        """Set baud rate."""
        self._obj.setBaudRate(baud_rate)

    def baudRate(self) -> int:
        """Return actual baud rate."""
        return self._obj.baudRate()

    def setDataBits(self, data_bits: "QtSerialPort.QSerialPort.DataBits") -> None:
        """Set data bits."""
        self._obj.setDataBits(data_bits)

    def dataBits(self) -> "QtSerialPort.QSerialPort.DataBits":
        """Return actual data bits."""
        return self._obj.dataBits()

    def setParity(self, parity: "QtSerialPort.QSerialPort.Parity") -> None:
        """Set parity check value."""
        self._obj.setParity(parity)

    def parity(self) -> "QtSerialPort.QSerialPort.Parity":
        """Return parity check."""
        return self._obj.parity()

    def setStopBits(self, stop_bit: "QtSerialPort.QSerialPort.StopBits") -> None:
        """Set stop bits."""
        self._obj.setStopBits(stop_bit)

    def stopBits(self) -> "QtSerialPort.QSerialPort.StopBits":
        """Return stop bits."""
        return self._obj.stopBits()

    def setFlowControl(self, flow: "QtSerialPort.QSerialPort.FlowControl") -> None:
        """Set flow conrol."""
        self._obj.setFlowControl(flow)

    def flowControl(self) -> "QtSerialPort.QSerialPort.FlowControl":
        """Return flow control."""
        return self._obj.flowControl()

    def setTimeOut(self, sec: int = 0, milisec: int = 3000) -> None:
        """Set time out."""
        time = milisec
        if sec:
            time = sec * 1000 + time

        self._obj.waitForBytesWritten(time)
        self._obj.waitForReadyRead(time)

    def open(self) -> bool:
        """Return if port is open."""
        return self._obj.open(QtCore.QIODeviceBase.OpenModeFlag.ReadWrite)

    def close(self) -> bool:
        """Return if port is closed."""
        self._obj.close()
        return True

    @decorators.not_implemented_warn
    def writeText(self, data: str) -> None:
        """Send string data."""
        pass

    @decorators.not_implemented_warn
    def getch(self) -> int:
        """Return int char readed from port."""
        return 0

    @decorators.not_implemented_warn
    def putch(self, char_: int) -> int:
        """Send a char."""
        return 0

    @decorators.not_implemented_warn
    def ungetch(self, char_: int) -> int:
        """Return unsigned char?."""
        return 0

    def size(self) -> int:
        """Return size data recceived."""
        return self._obj.size()

    def flush(self) -> bool:
        """Flush data."""
        return self._obj.flush()

    @decorators.not_implemented_warn
    def readBlock(self) -> int:
        """Read data block."""
        return 0

    def writeBlock(self) -> int:
        """Write data block."""
        return 0

    def bytesAvailable(self) -> int:
        """Return number of bytes avalaible to read."""
        return 0

    def setRts(self, value: bool) -> None:
        """Emit RTS. signal."""
        self._obj.setRequestToSend(value)

    def setDtr(self, value: bool) -> None:
        """Emit DTR signal."""
        self._obj.setDataTerminalReady(value)



"""Flpixmapview module."""
# -*- coding: utf-8 -*-
from PyQt6 import QtCore, QtWidgets, QtGui  # type: ignore
from typing import cast, Optional


class FLPixmapView(QtWidgets.QScrollArea):
    """FLPixmapView class."""

    # frame_ = None
    # scrollView = None
    _auto_scaled: bool
    _path: str
    _pixmap: QtGui.QPixmap
    _pixmapview: QtWidgets.QLabel
    _lay: QtWidgets.QHBoxLayout
    # gB_ = None
    _parent: QtWidgets.QWidget

    def __init__(self, parent: QtWidgets.QWidget) -> None:
        """Inicialize."""

        super(FLPixmapView, self).__init__(parent)
        self._auto_scaled = False
        self._lay = QtWidgets.QHBoxLayout(self)
        self._lay.setContentsMargins(0, 2, 0, 2)
        self._pixmap = QtGui.QPixmap()
        self._pixmapview = QtWidgets.QLabel(self)
        self._lay.addWidget(self._pixmapview)
        self._pixmapview.setAlignment(
            cast(
                QtCore.Qt.AlignmentFlag,
                QtCore.Qt.AlignmentFlag.AlignHCenter | QtCore.Qt.AlignmentFlag.AlignCenter,
            )
        )
        self._pixmapview.installEventFilter(self)
        self.setStyleSheet("QScrollArea { border: 1px solid darkgray; border-radius: 3px; }")
        self._parent = parent

    def setPixmap(self, pix: QtGui.QPixmap) -> None:
        """Set pixmap to object."""
        # if not project.DGI.localDesktop():
        #    project.DGI._par.addQueque("%s_setPixmap" % self._parent.objectName(
        #    ), self._parent.cursor_.valueBuffer(self._parent.fieldName_))
        #    return

        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.CursorShape.WaitCursor)
        self._pixmap = pix
        if self._pixmapview is not None:
            self._pixmapview.clear()
            self._pixmapview.setPixmap(self._pixmap)
        self.repaint()
        QtWidgets.QApplication.restoreOverrideCursor()

    def eventFilter(
        self, obj: Optional["QtCore.QObject"], event: Optional["QtCore.QEvent"]
    ) -> bool:
        """Event filter process."""

        if isinstance(obj, QtWidgets.QLabel) and isinstance(event, QtGui.QResizeEvent):
            self.resizeContents()

        return super().eventFilter(obj, event)  # type: ignore [arg-type]

    def resizeContents(self) -> None:
        """Resize contents to actual control size."""

        if self._pixmap is None or self._pixmap.isNull():
            return

        new_pix = self._pixmap
        if (
            self._auto_scaled is not None
            and self._pixmap is not None
            and self._pixmapview is not None
        ):
            if (
                self._pixmap.height() > self._pixmapview.height()
                or self._pixmap.width() > self._pixmapview.width()
            ):
                new_pix = self._pixmap.scaled(
                    self._pixmapview.size(), QtCore.Qt.AspectRatioMode.KeepAspectRatio
                )

            elif (
                self._pixmap.height() < self._pixmapview.pixmap().height()
                or self._pixmap.width() < self._pixmapview.pixmap().width()
            ):
                new_pix = self._pixmap.scaled(
                    self._pixmapview.size(), QtCore.Qt.AspectRatioMode.KeepAspectRatio
                )

        if self._pixmapview is not None:
            self._pixmapview.clear()
            self._pixmapview.setPixmap(new_pix)

    def previewUrl(self, url: str) -> None:
        """Set image from url."""

        qurl = QtCore.QUrl(url)
        if qurl.isLocalFile():
            path = qurl.path()

        if not path == self._path:
            self._path = path
            img = QtGui.QImage(self._path)

            if img is None:
                return

            pix = QtGui.QPixmap()
            QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.CursorShape.WaitCursor)
            pix.convertFromImage(img)
            QtWidgets.QApplication.restoreOverrideCursor()

            if pix is not None:
                self.setPixmap(pix)

    def clear(self) -> None:
        """Clear image into object."""

        if self._pixmapview is not None:
            self._pixmapview.clear()

    def pixmap(self) -> QtGui.QPixmap:
        """Return pixmap stored."""

        return self._pixmap

    def setAutoScaled(self, auto_scaled: bool) -> None:
        """Set auto sclate to the control."""

        self._auto_scaled = auto_scaled



"""Flfastcgi module."""

from pineboolib.core import decorators


class FLFastCgi(object):
    """FLFastCgi class."""

    @decorators.not_implemented_warn
    def getFullEnv(self) -> str:
        """Return full enviroment."""

        return ""

    @decorators.not_implemented_warn
    def getEnv(self, data: str) -> str:
        """Return enviroment."""

        return ""

    @decorators.not_implemented_warn
    def urlencode(self, data: str) -> str:
        """Return url encode."""

        return ""

    @decorators.not_implemented_warn
    def urldecode(self, data: str) -> str:
        """Return url decode."""

        return ""

    @decorators.not_implemented_warn
    def xmlentityencode(self, data: str) -> str:
        """Return xml entity encode."""

        return ""

    @decorators.not_implemented_warn
    def read(self, block_size: int = 65535) -> str:
        """Return xml entity decode."""

        return ""

    @decorators.not_implemented_warn
    def write(self, data: str) -> int:
        """Return xml entity decode."""

        return 0

    @decorators.not_implemented_warn
    def writeError(self, data: str) -> int:
        """Return error."""

        return 0



"""
Finalize pineboo setup and load.
"""

from pineboolib import logging
from typing import Optional, TYPE_CHECKING


from pineboolib.core import settings
from pineboolib.loader.preload import preload_actions

if TYPE_CHECKING:
    from PyQt6 import QtWidgets  # pragma: no cover
    from pineboolib.application import projectmodule  # pragma: no cover
    from pineboolib.interfaces import imainwindow, dgi_schema  # pragma: no cover
    from optparse import Values  # pragma: no cover


LOGGER = logging.get_logger(__name__)


def init_project(
    dgi: "dgi_schema.dgi_schema",
    options: "Values",
    project: "projectmodule.Project",
    main_window: Optional["imainwindow.IMainWindow"],
    app: "QtWidgets.QApplication",
) -> int:
    """Initialize the project and start it."""
    # from PyQt6 import QtCore  # type: ignore

    # if dgi.useDesktop() and dgi.localDesktop() and splash:
    #     splash.showMessage("Iniciando proyecto ...", QtCore.Qt.AlignLeft, QtCore.Qt.white)
    #     dgi.processEvents()

    project.message_manager().send("splash", "showMessage", ["Iniciando proyecto ..."])

    if options.preload:
        preload_actions(project, options.forceload)

        LOGGER.info("Finished preloading")
        return 0

    if "plus_sys" in project.actions.keys():
        project.call("formplus_sys.main", [], None, True)
        if options.quit_after_call:
            return 0

    call_function = settings.SETTINGS.value("application/callFunction", None)
    if options.call_function:
        call_function = options.call_function

    if call_function:
        LOGGER.debug("callFunction (%s)", call_function)
        args = call_function.split(":")
        try:
            project.call(args[0], args[1:] if len(args) > 1 else [])
        except Exception as error:
            from pineboolib import application

            LOGGER.warning("could not be executed %s : %s", call_function, str(error))

        if options.quit_after_call:
            return 0

    if main_window is not None:
        project.message_manager().send("splash", "showMessage", ["Creando interface ..."])
        # main_window = main_form.mainWindow
        main_window.initScript()
        ret = 0

        project.message_manager().send("splash", "showMessage", ["Abriendo interfaz ..."])
        main_window.show()
        project.message_manager().send("splash", "showMessage", ["Listo ..."])
        project.message_manager().send("splash", "hide")
    # FIXME: Is always None because the earlier code is commented out
    # if objaction:
    #     project.openDefaultForm(objaction.form())

    ret = app.exec() if dgi.localDesktop() else dgi.exec()

    # if main_form is not None:
    #    main_form.mainWindow = None
    #    del main_window
    del project
    return ret



"""Qtoolbutton module."""
# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore[import]
from pineboolib.core import decorators


from pineboolib.q3widgets.qframe import QFrame
from pineboolib.q3widgets.qgroupbox import QGroupBox
from pineboolib.q3widgets.qwidget import QWidget
from pineboolib.q3widgets.qiconset import QIconSet

from typing import Union, Optional


class QToolButton(QtWidgets.QToolButton):
    """QToolButton class."""

    groupId: Optional[int]

    def __init__(
        self, parent: Union[QWidget, QGroupBox, QFrame], name: Optional[str] = None
    ) -> None:
        """Inicialize."""
        super().__init__(parent)

        if name is not None:
            self.setObjectName(name)

        self.groupId = None  # pylint: disable=invalid-name

    def setToggleButton(self, value: bool) -> None:
        """Set toggled button."""

        self.setDown(value)

    @decorators.deprecated
    def setUsesBigPixmap(self, value: bool):
        """Set uses big pixmap."""

        pass

    def toggleButton(self) -> bool:
        """Return button is toggled."""
        return self.isDown()

    def getOn(self) -> bool:
        """Return button is checked."""
        return self.isChecked()

    def setOn(self, value: bool) -> None:
        """Set checked."""
        self.setChecked(value)

    @decorators.deprecated
    def setUsesTextLabel(self, value: bool):
        """Set uses text label."""
        pass

    def buttonGroupId(self) -> Optional[int]:
        """Return button group id."""
        return self.groupId

    def setButtonGroupId(self, id_: int) -> None:
        """Set button group id."""
        self.groupId = id_

    def getEnabled(self) -> bool:
        """Return if is enabled."""

        return super().isEnabled()

    def setIconSet(self, icon_set: "QIconSet") -> None:
        """Set iconset."""

        self.setIcon(icon_set)

    def setEnabled(self, value: bool) -> None:
        """Set if enabled."""

        super().setEnabled(value)

    def animateClick(self, num: int) -> None:  # type: ignore [override] # noqa: F821
        """Animateclick Bound method."""

        super().animateClick()

    on = property(getOn, setOn)
    enabled = property(getEnabled, setEnabled)



"""Qtabwidget module."""
# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore[import]
from pineboolib import logging
from typing import Optional, Union

logger = logging.get_logger(__name__)


class QTabWidget(QtWidgets.QTabWidget):
    """QTabWidget class."""

    Top = QtWidgets.QTabWidget.TabPosition.North
    Bottom = QtWidgets.QTabWidget.TabPosition.South
    Left = QtWidgets.QTabWidget.TabPosition.West
    Right = QtWidgets.QTabWidget.TabPosition.East

    def setTabEnabled(self, tab: str, enabled: bool) -> None:  # type: ignore [override]
        """Set a tab enabled."""
        idx = self.indexByName(tab)
        if idx is None:
            return None

        QtWidgets.QTabWidget.setTabEnabled(self, idx, enabled)

    def showPage(self, tab: str) -> None:
        """Show a tab specified by name."""
        idx = self.indexByName(tab)
        if idx is None:
            return None

        QtWidgets.QTabWidget.setCurrentIndex(self, idx)

    def indexByName(self, tab: Union[str, int]) -> Optional[int]:
        """Return a index tab from a name or number."""
        if isinstance(tab, int):
            return tab
        elif not isinstance(tab, str):
            logger.error("ERROR: Unknown type tab name or index:: QTabWidget %r", tab)
            return None

        try:
            for num in range(self.count()):
                if self.widget(num).objectName() == tab.lower():  # type: ignore [union-attr]
                    return num
        except ValueError:
            logger.error("ERROR: Tab not found:: QTabWidget, tab name = %r", tab)
        return None

    def removePage(self, idx) -> None:
        """Remove a page specified by name."""

        if isinstance(idx, int):
            self.removeTab(idx)



"""Test_xmlaction module."""
import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application


class TestXMLAction(unittest.TestCase):
    """TestXMLAction Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Test basic."""

        action = application.PROJECT.actions["flareas"]
        self.assertEqual(action._name, "flareas")
        self.assertEqual(action._table, "flareas")
        self.assertEqual(action._master_form, "master")
        self.assertEqual(action._record_form, "flareas")
        self.assertEqual(action._master_script, "flmasterareas")
        self.assertEqual(action._record_script, "")

        action2 = application.PROJECT.actions["flmodules"]
        self.assertEqual(action2._name, "flmodules")
        self.assertEqual(action2._table, "flmodules")
        self.assertEqual(action2._master_form, "master")
        self.assertEqual(action2._record_form, "flmodulos")
        self.assertEqual(action2._master_script, "")
        self.assertEqual(action2._record_script, "flmodules")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""
Replacement for Python logging that adds trace and other methods.

It allows MyPy/PyType to properly keep track of the new message types
"""

import logging as python_logging
from logging import handlers
from logging import basicConfig  # noqa: F401
from typing import Any, Set, Optional
import os
import sys
import gzip

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
MESSAGE = 25  # NEW
INFO = 20
HINT = 18  # NEW
NOTICE = 15  # NEW
DEBUG = 10
TRACE = 5  # NEW
NOTSET = 0

log_file_dir = (
    "/var/log" if not sys.platform.startswith("win") else os.path.join(os.environ["ProgramFiles"])
)


if not os.access(log_file_dir, os.W_OK):
    log_file_dir = os.path.join(os.path.expanduser("~"), "Pineboo", "log")
else:
    log_file_dir = os.path.join(log_file_dir, "Pineboo")

file_dir: Optional[str] = os.environ.get("PINEBOODIR")
if file_dir:
    log_file_dir = os.path.join("/pineboo/pineboo", "log")

LOG_FILE_PATH: str = os.path.join(log_file_dir, "pineboo.log")
LOG_FILE_BACKUP_COUNTS: int = 30  # ficheros de backup
LOG_FILE_FORMAT: str = "%(asctime)s - %(process)d - %(name)s - %(levelname)s - %(message)s"


class GZipRotator:
    def __call__(self, source: str, dest: str) -> None:
        os.rename(source, dest)
        f_in = open(dest, "rb")
        f_out = gzip.open("%s.gz" % dest, "wb")
        f_out.writelines(f_in)
        f_out.close()
        f_in.close()
        os.remove(dest)


class Logger(python_logging.Logger):
    """
    Replaced Logger object.

    Adds message, hint, notice  and trace
    """

    PINEBOO_DEFAULT_LEVEL = 0
    PINEBOO_LOGGERS: Set["Logger"] = set()

    @classmethod
    def set_pineboo_default_level(cls, level: int) -> None:
        """Set the default logging level for the whole app."""
        cls.PINEBOO_DEFAULT_LEVEL = level
        for logger in cls.PINEBOO_LOGGERS:
            logger.setLevel(level)

    def message(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Log a message."""
        self.log(MESSAGE, message, *args, **kwargs)

    def hint(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Log a hint."""
        self.log(HINT, message, *args, **kwargs)

    def notice(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Log a notice."""
        self.log(NOTICE, message, *args, **kwargs)

    def trace(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Log a trace."""
        self.log(TRACE, message, *args, **kwargs)


python_logging.Logger.manager.loggerClass = Logger  # type: ignore


def get_logger(name: Optional[str] = None) -> Logger:
    """
    Return a logger with the specified name, creating it if necessary.

    If no name is specified, return the root logger.
    """

    if name:
        # print(name)
        # if not name.startswith("pineboolib."):
        #     raise ValueError("Invalid logger name %r, should be the module path.")
        logger: Logger = python_logging.Logger.manager.getLogger(name)  # type: ignore
        Logger.PINEBOO_LOGGERS.add(logger)
        if Logger.PINEBOO_DEFAULT_LEVEL != 0 and logger.level == 0:
            logger.setLevel(Logger.PINEBOO_DEFAULT_LEVEL)

        can_log_to_file = True
        base_dir = os.path.dirname(LOG_FILE_PATH)
        if not os.path.exists(base_dir):
            try:
                os.makedirs(os.path.dirname(LOG_FILE_PATH), exist_ok=True)
            except Exception:

                can_log_to_file = False
        if can_log_to_file:
            can_log_to_file = os.access(base_dir, os.W_OK)

        if not can_log_to_file:
            logger.error("* Cannot write to log file %s (%s)" % (LOG_FILE_PATH, name))
        else:
            file_handler = handlers.TimedRotatingFileHandler(
                LOG_FILE_PATH, backupCount=LOG_FILE_BACKUP_COUNTS, when="midnight"
            )
            file_handler.setLevel(INFO)
            file_handler.setFormatter(python_logging.Formatter(LOG_FILE_FORMAT))
            file_handler.rotator = (
                GZipRotator()
            )  # https://stackoverflow.com/questions/8467978/python-want-logging-with-log-rotation-and-compression
            logger.addHandler(file_handler)

        return logger
    else:
        raise Exception("Pineboo getLogger does not allow for root logger")


def _add_logging_level(level_name: str, level_num: int) -> None:
    method_name = level_name.lower()

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def log_for_level(self: Any, message: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    python_logging.addLevelName(level_num, level_name)
    setattr(python_logging, level_name, level_num)
    if not hasattr(python_logging.getLoggerClass(), method_name):
        setattr(python_logging.getLoggerClass(), method_name, log_for_level)


_add_logging_level("TRACE", TRACE)
_add_logging_level("NOTICE", NOTICE)
_add_logging_level("HINT", HINT)
_add_logging_level("MESSAGE", MESSAGE)



"""Flex module."""

import ply.lex as lex  # type: ignore
from pineboolib.application.parsers.parser_qsa import token_rules

from pineboolib import logging


# TODO: Cada vez que se cambia este fichero, se tiene que lanzar sin el "-OO" de python para acelerar. Construye entonces
# ..... el fichero de cache que subimos a git, y se relee desde ahí las siguientes veces con el -OO.
# ..... Si da problemas, hay que volver a optimize=0 y/o eliminar lextab.py
lexer = lex.lex(debug=False, optimize=1, module=token_rules, debuglog=logging)
if __name__ == "__main__":
    lex.runmain(lexer)



"""Test_aqsutil module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.fllegacy.aqsobjects import aqutil


class TestAQUtil(unittest.TestCase):
    """TestAQUtil Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_nombreCampos(self) -> None:
        """Return array from nombreCampos."""
        from pineboolib.qsa import qsa

        util = aqutil.AQUtil()
        array = util.nombreCampos("flmodules")
        self.assertTrue(qsa.parseInt(array[0]))
        value = array.pop(0)
        self.assertEqual(value, "6")
        self.assertEqual(
            str(array),
            str(qsa.Array("bloqueo", "idmodulo", "idarea", "descripcion", "version", "icono")),
        )

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Manage cached xpm."""

import os
import os.path

from pineboolib.core import settings
from pineboolib import logging, application

LOGGER = logging.get_logger(__name__)


def cache_xpm(value: str, xpm_name: str = "") -> str:
    """
    Return a path to a file with the content of the specified string.

    @param value. text string with the xpm or path to this.
    @return file path contains Xpm
    """

    if not value:
        LOGGER.warning("the value is empty!")
        return ""

    if not xpm_name:
        xpm_name = value[: value.find("[]")]
        xpm_name = xpm_name[xpm_name.rfind(" ") + 1 :]

    conn = application.PROJECT.conn_manager.mainConn()
    if conn is None:
        raise Exception("Project is not connected yet")

    cache_dir = "%s/cache/%s/cacheXPM" % (application.PROJECT.tmpdir, conn.DBName())
    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)

    file_name = (
        "%s/%s" % (cache_dir, value[value.find("cacheXPM") + 9 :])
        if value.find("cacheXPM") > -1
        else "%s/%s.xpm" % (cache_dir, xpm_name)
    )

    if not os.path.exists(file_name) or settings.CONFIG.value(
        "ebcomportamiento/no_img_cached", False
    ):
        file_ = open(file_name, "w")
        file_.write(value)
        file_.close()

    return file_name



"""Qdateedit module."""
# -*- coding: utf-8 -*-

from PyQt6 import QtWidgets, QtCore  # type: ignore[import]
from pineboolib.core import decorators
from typing import Any, Union


class QDateEdit(QtWidgets.QDateEdit):
    """QDateEdit class."""

    _parent: QtWidgets.QWidget
    _date: str
    separator_ = "-"

    def __init__(self, parent=None, name=None) -> None:
        """Inicialize."""

        super().__init__(parent)
        super().setDisplayFormat("dd-MM-yyyy")
        if name:
            self.setObjectName(name)
        self.setSeparator("-")
        self._parent = parent
        self.date_ = super().date().toString(QtCore.Qt.DateFormat.ISODate)
        # if not project.DGI.localDesktop():
        #    project.DGI._par.addQueque("%s_CreateWidget" % self._parent.objectName(), "QDateEdit")

    def getDate(self) -> str:
        """Return string date."""
        ret = super().date().toString(QtCore.Qt.DateFormat.ISODate)
        if ret != "2000-01-01":
            return ret
        else:
            return ""

    def setDate(self, value: Union[str, Any]) -> None:
        """Set date."""

        if not isinstance(value, str):
            if hasattr(value, "toString"):
                value = value.toString("yyyy%sMM%sdd" % (self.separator(), self.separator()))

            value = str(value)

        date = QtCore.QDate.fromString(value[:10], "yyyy-MM-dd")
        super().setDate(date)
        # if not project.DGI.localDesktop():
        #    project.DGI._par.addQueque("%s_setDate" % self._parent.objectName(), "QDateEdit")

    date: str = property(getDate, setDate)  # type: ignore[assignment] # noqa : F821

    @decorators.not_implemented_warn
    def setAutoAdvance(self, value: bool) -> None:
        """Set auto advance."""
        pass

    def setSeparator(self, value: str) -> None:
        """Set separator."""

        self.separator_ = value
        self.setDisplayFormat("dd%sMM%syyyy" % (self.separator(), self.separator()))

    def separator(self) -> str:
        """Return separator."""

        return self.separator_

    # def __getattr__(self, name) -> Any:
    #    """Return attribute."""
    #
    #    if name == "date":
    #        return super(QDateEdit, self).date().toString(QtCore.Qt.ISODate)



"""Qlistview module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtCore, QtWidgets, QtGui  # type: ignore[import]

from typing import Any, List, Optional, Union, cast


class QListView(QtWidgets.QWidget):
    """QListView class."""

    _resizeable: bool
    _clickable: bool
    _root_is_decorated: bool
    _default_rename_action: bool
    _tree: QtWidgets.QTreeView
    _cols_labels: List[str]
    _key: str
    _root_item: Any
    _current_row: int

    doubleClicked = QtCore.pyqtSignal(object)
    selectionChanged = QtCore.pyqtSignal(object)
    expanded = QtCore.pyqtSignal(object)
    collapsed = QtCore.pyqtSignal(object)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Inicialize."""

        super().__init__(parent=None)
        lay = QtWidgets.QVBoxLayout(self)
        self._tree = QtWidgets.QTreeView(self)
        lay.addWidget(self._tree)
        self._tree.setModel(QtGui.QStandardItemModel())
        self._cols_labels = []
        self._resizeable = True
        self._clickable = True
        self._default_rename_action = False
        self._root_is_decorated = False
        self._key = ""
        self._root_item = None
        self._current_row = -1
        cast(
            QtCore.pyqtSignal, self._tree.doubleClicked
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self.doubleClickedEmit
        )
        cast(
            QtCore.pyqtSignal, self._tree.clicked
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self.singleClickedEmit
        )
        cast(
            QtCore.pyqtSignal, self._tree.activated
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self.singleClickedEmit
        )

    def singleClickedEmit(self, index: Any) -> None:
        """Emit single clicked signal."""
        if not self._clickable:
            return

        if index.column() != 0:
            index = index.sibling(index.row(), 0)
        else:
            index = index.sibling(index.row(), index.column())
        item = index.model().itemFromIndex(index)

        self.selectionChanged.emit(item)

    def doubleClickedEmit(self, index: Any) -> None:
        """Emit double clicked signal."""
        if not self._clickable:
            return

        item = index.model().itemFromIndex(index)
        self.doubleClicked.emit(item)

    def addItem(self, item_text: str) -> None:
        """Add a new item."""

        from pineboolib.fllegacy.fllistviewitem import FLListViewItem

        self._current_row = self._current_row + 1
        item = FLListViewItem()
        item.setEditable(False)
        item.setText(item_text)
        if self._tree is not None:
            cast(QtGui.QStandardItemModel, self._tree.model()).setItem(self._current_row, 0, item)

    def setHeaderLabel(self, labels: Union[str, List[str]]) -> None:
        """Set header labels from a stringlist."""
        if isinstance(labels, str):
            labels = [labels]

        if self._tree is not None:
            cast(QtGui.QStandardItemModel, self._tree.model()).setHorizontalHeaderLabels(labels)
        self._cols_labels = labels

    def setColumnText(self, col: int, new_value: str) -> None:
        """Set Column text."""

        i = 0
        new_list = []
        for old_value in self._cols_labels:
            value = new_value if i == col else old_value
            new_list.append(value)

        self._cols_labels = new_list

    def addColumn(self, text: str) -> None:
        """Add a new column."""

        self._cols_labels.append(text)

        self.setHeaderLabel(self._cols_labels)

    def setClickable(self, clickable: bool) -> None:
        """Set clickable."""
        self._clickable = clickable

    def setResizable(self, resizeable: bool) -> None:
        """Set resizeable."""
        self._resizeable = resizeable

    def eventFilter(
        self, obj: Optional["QtCore.QObject"], event: Optional["QtCore.QEvent"]
    ) -> bool:
        """Event filter."""

        if isinstance(event, QtGui.QResizeEvent):
            if not self._resizeable:
                return False

        return super().eventFilter(obj, event)  # type: ignore [arg-type]

    def setItemMargin(self, item_margin: int) -> None:
        """Set items margin."""

        style_ = "QTreeView::item#%s { border: 0px; padding: %spx; }" % (
            self.objectName(),
            item_margin,
        )
        self._tree.setStyleSheet(style_)

    # @decorators.not_implemented_warn
    # def resizeEvent(self, e: QtCore.QEvent) -> None:
    #    """Process resize event."""
    #    if self._resizeable:
    #        super().resizeEvent(e)

    def clear(self) -> None:
        """Clear all data."""

        self._cols_labels = []

    def getDefaultRenameAction(self) -> bool:
        """Return default_rename_action enabled."""
        return self._default_rename_action

    def setDefaultRenameAction(self, default: bool) -> None:
        """Set default_rename_action enabled."""
        self._default_rename_action = default

    def model(self) -> QtGui.QStandardItemModel:
        """Return model index."""

        if self._tree is not None:
            return cast(QtGui.QStandardItemModel, self._tree.model())
        else:
            raise Exception("No hay _tree")

    def setRootIsDecorated(self, show: bool) -> None:
        """Set tree root decorated."""

        self._tree.setRootIsDecorated(show)

    def getRootIsDecorated(self) -> bool:
        """Return if tree is root decorated."""

        return self._tree.rootIsDecorated()

    rootIsDecorated = property(getRootIsDecorated, setRootIsDecorated)
    defaultRenameAction = property(getDefaultRenameAction, setDefaultRenameAction)



# -*- coding: utf-8 -*-
"""
AQSql Module.

Provide queries to DB.
"""
from enum import IntEnum, unique, Enum

from pineboolib import application, logging
from pineboolib.application import types
from pineboolib.application.database import pnsqlcursor

from typing import Union, Any, List, TYPE_CHECKING


if TYPE_CHECKING:
    from pineboolib.interfaces import iconnection, isqlcursor  # pragma: no cover

LOGGER = logging.get_logger(__name__)


class TableType(Enum):
    """TableType class."""

    Tables: int = 1  # pylint: disable=invalid-name
    SystemTables: int = 2  # pylint: disable=invalid-name
    Views: int = 3  # pylint: disable=invalid-name
    AllTables: int = 0  # pylint: disable=invalid-name


class AQSql(object):
    """AQSql Class."""

    Insert: int = 0
    Edit: int = 1
    Del: int = 2
    Browse: int = 3

    @unique
    class TableType(IntEnum):
        """TableType class."""

        Tables = 0x01  # pylint: disable=invalid-name
        SystemTables = 0x02  # pylint: disable=invalid-name
        Views = 0x04  # pylint: disable=invalid-name
        AllTables = 0xFF  # pylint: disable=invalid-name

    @classmethod
    def database(self, connection_name: str = "default") -> "iconnection.IConnection":
        """Return the database of a connection."""

        if application.PROJECT.conn_manager is None:
            raise Exception("Project is not connected yet")
        return application.PROJECT.conn_manager.useConn(connection_name)

    @classmethod
    def update(
        self,
        table_or_cursor: Union[str, "isqlcursor.ISqlCursor"],
        fields: Union[List[str], types.Array],
        values: Union[List[Any], types.Array],
        where: str = "",
        conn: str = "default",
    ):
        """Update a set of cursor records with new values."""
        cur: "isqlcursor.ISqlCursor"

        if isinstance(table_or_cursor, str):
            cur = pnsqlcursor.PNSqlCursor(table_or_cursor, conn)
        else:
            cur = table_or_cursor

        # if cur is None:
        #     return False

        if not cur.metadata():
            return False

        cur.select(where)

        ret_ = True
        activate_check_integrity = cur.activatedCheckIntegrity()

        fields_count = len(fields)

        while ret_ and cur.next():
            cur.setModeAccess(cur.Edit)
            cur.refreshBuffer()

            for i in range(0, fields_count - 1):
                cur.setValueBuffer(fields[i], values[i])

            msg_check_integrity = cur.msgCheckIntegrity()
            if msg_check_integrity != "":
                raise Exception(msg_check_integrity)

            cur.setActivatedCheckIntegrity(False)
            ret_ = cur.commitBuffer()
            cur.setActivatedCheckIntegrity(activate_check_integrity)

        return ret_

    @classmethod
    def insert(
        self,
        table_or_cursor: Union[str, "isqlcursor.ISqlCursor"],
        fields: Union[List[str], types.Array],
        values: Union[List[Any], types.Array],
        conn: str = "default",
    ):
        """Insert a record in a cursor."""
        cur: "isqlcursor.ISqlCursor"

        cur = (
            pnsqlcursor.PNSqlCursor(table_or_cursor, conn)
            if isinstance(table_or_cursor, str)
            else table_or_cursor
        )

        if not cur.metadata():
            return False

        fields_count = len(fields)

        cur.setModeAccess(cur.Insert)
        cur.refreshBuffer()

        for i in range(fields_count):
            cur.setValueBuffer(fields[i], values[i])

        msg_check_integrity = cur.msgCheckIntegrity()
        if msg_check_integrity != "":
            raise Exception(msg_check_integrity)

        ret_ = False
        activated_check_integrity = cur.activatedCheckIntegrity()
        cur.setActivatedCheckIntegrity(False)
        ret_ = cur.commitBuffer()
        cur.setActivatedCheckIntegrity(activated_check_integrity)

        return ret_

    @classmethod
    def del_(
        self,
        cur_or_table: Union[str, "isqlcursor.ISqlCursor"],
        where: str = "",
        conn_name: str = "default",
    ):
        """Remove a recordset from a cursor."""

        cur: "isqlcursor.ISqlCursor"

        if not isinstance(cur_or_table, str):
            cur = cur_or_table
        else:
            cur = pnsqlcursor.PNSqlCursor(cur_or_table, True, conn_name)

        act_check = cur.activatedCheckIntegrity()

        cur.select(where)
        if cur.first():
            while True:
                cur.setModeAccess(cur.Del)
                cur.refreshBuffer()
                cur.setActivatedCheckIntegrity(False)
                res_commit = cur.commitBuffer()
                cur.setActivatedCheckIntegrity(act_check)
                if not res_commit:
                    return False

                if not cur.next():
                    break

        return True


"""
FIXME!!: crear todos los emun...
Código C++ original
class AQSql : public QObject
{
  Q_OBJECT

  Q_ENUMS(ModeAccess)
  Q_ENUMS(ConnOptions)
  Q_ENUMS(SpecialType)
  Q_ENUMS(Cardinality)
  Q_ENUMS(TableType)
  Q_ENUMS(SqlErrorType)
  Q_ENUMS(Refresh)

public:
  enum ModeAccess {
    Insert = 0,
    Edit = 1,
    Del = 2,
    Browse = 3
  };

  enum ConnOptions {
    User = 0,
    Password = 1,
    Port = 2,
    Host = 3
  };

  enum SpecialType {
    Serial = 100,
    Unlock = 200,
    Check = 300
  };

  enum Cardinality {
    RELATION_1M = 0,
    RELATION_M1 = 1
  };

  enum SqlErrorType {
    SqlErrorNone,
    SqlErrorConnection,
    SqlErrorStatement,
    SqlErrorTransaction,
    SqlErrorUnknown
  };

  enum Refresh {
    RefreshData = 1,
    RefreshColumns = 2,
    RefreshAll = 3
  };

  AQSql() : QObject(0, "aqs_aqsql") {}

  void throwError(const QString &msg, FLSqlDatabase *db) {
    if (!db->qsaExceptions() || !globalAQSInterpreter) {
      qWarning(msg);
      return;
    }
    if (globalAQSInterpreter->isRunning())
      globalAQSInterpreter->throwError(msg);
    else
      qWarning(msg);
  }

public slots:
  /**
  Añade una base de datos a las conexiones disponibles.

  La base de datos será abierta. Si ya existiera una conexión con el mismo nombre
  la base datos correspondiente será cerrada y borrada, sustituyéndola por la nueva.

  @param driverAlias Alias del driver ( PostgreSQL, MySQL, SQLite, ... ), ver FLSqlDatabase.
  @param nameDB  Nombre de la base de datos a la que conectar
  @param user  Usuario de la conexión
  @param password Contraseña para el usuario
  @param host  Nombre o dirección del servidor de la base de datos
  @param port  Puerto TCP de conexion
  @param connectionName Nombre de la conexion
    @param connectOptions Contiene opciones auxiliares de conexión a la base de datos.
                        El formato de la cadena de opciones es una lista separada por punto y coma
                        de nombres de opción o la opción = valor. Las opciones dependen del uso del
                        driver de base de datos.
  @return TRUE si se pudo realizar la conexión, FALSE en caso contrario
  */
  bool addDatabase(const QString &driverAlias, const QString &nameDB, const QString &user,
                   const QString &password, const QString &host, int port,
                   const QString &connectionName, const QString &connectOptions = QString::null) {
    return FLSqlConnections::addDatabase(driverAlias, nameDB, user, password, host, port,
                                         connectionName, connectOptions);
  }

  /**
  Sobrecargada por conveniencia

  Practicamente hace lo mismo que el método anterior pero utilizando una base de datos ya construida

  @param db  Base datos a añadir a las conexiones disponibles, ver FLSqlDatabase.
  @param connectionName Nombre de la conexion
  @return TRUE si se pudo realizar la conexión, FALSE en caso contrario
  */
  bool addDatabase(FLSqlDatabase *db, const QString &connectionName = "default") {
    return FLSqlConnections::addDatabase(db, connectionName);
  }

  /**
  Sobrecargada por conveniencia

  Añade una base de datos a las conexiones disponibles utilizando los datos de otra conexión

  @param newConnName    Nombre a utilizar para la nueva conexion
  @param sourceConnName Nombre de una conexión existente a utilizar como origen de los datos de conexión
  @return TRUE si se pudo realizar la conexión, FALSE en caso contrario
  */
  bool addDatabase(const QString &newConnName, const QString &sourceConnName = "default") {
    FLSqlDatabase *srcDb = FLSqlConnections::database(sourceConnName);
    if (!srcDb)
      return false;
    return FLSqlConnections::addDatabase(srcDb->driverName(), srcDb->database(),
                                         srcDb->user(), srcDb->password(), srcDb->host(),
                                         srcDb->port(), newConnName);
  }

  /**
  Elimina una base de datos de las conexiones disponibles.

  Cierra la base de datos correspondiente y la elimina.

  @param connectionName Nombre de la conexion
  @return TRUE si se pudo eliminar la base de datos, FALSE en caso contrario
  */
  bool removeDatabase(const QString &connectionName) {
    return FLSqlConnections::removeDatabase(connectionName);
  }

  /**
  Obtiene la base de datos de una conexion.

  @param connectionNmae Nombre de la conexion
  @return La base de datos correspondiente al nombre de conexion indicado
  */
  FLSqlDatabase *database(const QString &connectionName = "default") {
    return FLSqlConnections::database(connectionName);
  }

  /**
  Finalizar todas las conexiones
  */
  void finish() {
    FLSqlConnections::finish();
  }

  /**
  Inserta un registro en un cursor

  Si hay un error SQL, eleva una excepción con el mensaje de error

  Ejemplo:

  var cur = new AQSqlCursor("clientes");
  try {
    AQSql.insert(cur,
                 ["codcliente","nombre","cifnif","codserie"],
                 ["1","pepe","XYZ","A"]);
  } catch (e) {
    sys.errorMsgBox("Error SQL: " + e);
  }
  */
  bool insert(FLSqlCursor *cur, const QStringList &fields,
              const QValueList<QVariant> &values) {
    if (!cur)
      return false;

    if (!cur->metadata()) {
      throwError(
        tr("No hay metadatos para '%1'").arg(cur->QObject::name()), cur->db()
      );
      return false;
    }

    int fields_count = fields.size();
    int valuesCount = values.size();

    cur->setModeAccess(Insert);
    if (!cur->refreshBuffer())
      return false;
    for (int i = 0; i < fields_count; ++i)
      cur->setValueBuffer(fields[i], (i < valuesCount ? values[i] : QVariant()));

    QString msgCheck(cur->msgCheckIntegrity());
    if (!msgCheck.isEmpty()) {
      throwError(msgCheck, cur->db());
      return false;
    }

    bool actCheck = cur->activatedCheckIntegrity();
    cur->setActivatedCheckIntegrity(false);
    bool ok = cur->commitBuffer();
    cur->setActivatedCheckIntegrity(actCheck);
    return ok;
  }

  bool insert(const QString &table, const QStringList &fields,
              const QValueList<QVariant> &values,
              const QString &connName = "default") {
    AQSqlCursor cur(table, true, connName);
    return insert(&cur, fields, values);
  }

  /**
  Actualiza un conjunto de registros de un cursor con nuevos valores

  Si hay un error SQL, eleva una excepción con el mensaje de error

  Ejemplo:

  var cur = new AQSqlCursor("clientes");
  try {
    AQSql.update(cur,
                 ["nombre","cifnif","codserie"],
                 ["juan","ZYX","A"],
                 "codcliente='1'");
  } catch (e) {
    sys.errorMsgBox("Error SQL: " + e);
  }
  */
  bool update(FLSqlCursor *cur, const QStringList &fields,
              const QValueList<QVariant> &values, const QString &where = "") {
    if (!cur)
      return false;

    if (!cur->metadata()) {
      throwError(
        tr("No hay metadatos para '%1'").arg(cur->QObject::name()), cur->db()
      );
      return false;
    }

    if (!cur->select(where))
      return false;

    bool ok = true;
    QString msgCheck;
    bool actCheck = cur->activatedCheckIntegrity();
    int fields_count = fields.size();
    int valuesCount = values.size();

    while (ok && cur->next()) {
      cur->setModeAccess(Edit);
      if (!cur->refreshBuffer()) {
        ok = false;
        break;
      }
      for (int i = 0; i < fields_count; ++i)
        cur->setValueBuffer(fields[i], (i < valuesCount ? values[i] : QVariant()));

      msgCheck = cur->msgCheckIntegrity();
      if (!msgCheck.isEmpty()) {
        ok = false;
        throwError(msgCheck, cur->db());
        break;
      }

      cur->setActivatedCheckIntegrity(false);
      ok = cur->commitBuffer();
      cur->setActivatedCheckIntegrity(actCheck);
    }
    return ok;
  }

  bool update(const QString &table, const QStringList &fields,
              const QValueList<QVariant> &values, const QString &where = "",
              const QString &connName = "default") {
    AQSqlCursor cur(table, true, connName);
    cur.setForwardOnly(true);
    return update(&cur, fields, values, where);
  }

  /**
  Elimina un conjunto de registros de un cursor

  Si hay un error SQL, eleva una excepción con el mensaje de error

  Ejemplo:

  var cur = new AQSqlCursor("clientes");
  try {
    AQSql.del(cur, "codcliente='1'");
  } catch (e) {
    sys.errorMsgBox("Error SQL: " + e);
  }
  */
  bool del(FLSqlCursor *cur, const QString &where = "") {
    if (!cur)
      return false;

    if (!cur->metadata()) {
      throwError(
        tr("No hay metadatos para '%1'").arg(cur->QObject::name()), cur->db()
      );
      return false;
    }

    if (!cur->select(where))
      return false;

    bool ok = true;
    QString msgCheck;
    bool actCheck = cur->activatedCheckIntegrity();

    while (ok && cur->next()) {
      cur->setModeAccess(Del);
      if (!cur->refreshBuffer()) {
        ok = false;
        break;
      }

      msgCheck = cur->msgCheckIntegrity();
      if (!msgCheck.isEmpty()) {
        ok = false;
        throwError(msgCheck, cur->db());
        break;
      }

      cur->setActivatedCheckIntegrity(false);
      ok = cur->commitBuffer();
      cur->setActivatedCheckIntegrity(actCheck);
    }
    return ok;
  }

  bool del(const QString &table, const QString &where = "",
           const QString &connName = "default") {
    AQSqlCursor cur(table, true, connName);
    cur.setForwardOnly(true);
    return del(&cur, where);
  }

  /**
  Ejecuta una consulta y devuelve información de la misma y el conjunto de
  registros obtenidos

  Si hay un error SQL, eleva una excepción con el mensaje de error

  Devuelve un array 'A' donde:

  -A[0] contiene el número de registros
  -A[1] contiene el número de campos
  -A[2] contiene otro array con los nombres de lo campos, en orden correlativo
        al que aparecen en la consulta
  -A[3..fin] los valores de los campos, en grupos de tamaño A[1]

  Ejemplo:

  var records;
  try {
    records = AQSql.select("bancos.*", "bancos");
    if (!records.length)
      return;

    var size = records[0];
    var nFields = records[1];
    var fieldNames = records[2];

    var rec = "";
    for (var i = 0; i < nFields; ++i)
      rec += fieldNames[i] + "    | ";
    print(rec);
    print("==============================================");

    for (var i = 3; i <= size * nFields;  i += nFields) {
      rec = "";
      for (var j = 0; j < nFields; ++j) {
        rec += records[i + j] + " | ";
      }
      print(rec);
    }
  } catch (e) {
    sys.errorMsgBox("Error SQL: " + e);
  }
  */
  QValueList<QVariant> select(const QString &select,
                              const QString &from,
                              const QString &where = QString::null,
                              const QString &orderBy = QString::null,
                              const QString &connName = "default") {
    AQSqlQuery qry(0, connName);
    qry.setTablesList(from);
    qry.setSelect(select);
    qry.setFrom(from);
    qry.setWhere(where);
    qry.setOrderBy(orderBy);
    qry.setForwardOnly(true);
    if (!qry.exec())
      return QVariantList();

    QVariantList ret;
    int countFields = qry.fieldList().count();

    ret.append(qry.size());
    ret.append(countFields);
    ret.append(qry.fieldList());

    while (qry.next()) {
      for (int i = 0; i < countFields; ++i)
        ret.append(qry.value(i));
    }

    return ret;
  }

  /**
  Esencialmente hace lo mismo que AQSql::select(), pero con la diferencia que reservará
  un bloqueo sobre los registros que devuelve la consulta.

  Si hay otro selectForUpdate anterior sobre los mismos o algunos registros que obtiene la
  consulta esta llamanda quedará en espera, bloqueada sobre esos registros, hasta que termine
  la transacción que inició el otro selectForUpdate.

  En PostgreSQL se puede utizar el parámetro 'nowait'. Si es TRUE y si al ejecutar la
  consulta se detecta que se va a caer en un bloqueo, es decir hay otro selectForUpdate anterior,
  no se bloqueará la llamada, y se elevará una excepción, terminando la transacción en curso.
  El parámetro 'nowait' no tiene efecto para cualquier otra base de datos distinta a PostgreSQL.
  */
  QValueList<QVariant> selectForUpdate(const QString &select,
                                       const QString &from,
                                       const QString &where = QString::null,
                                       bool nowait = false,
                                       const QString &connName = "default") {
    AQSqlQuery qry(0, connName);
    qry.setTablesList(from);
    qry.setSelect(select);
    qry.setFrom(from);
    QString w(where.isEmpty() ? "1=1" : w);
    qry.setWhere(nowait ?
                 w + QString::fromLatin1(" FOR UPDATE NOWAIT") :
                 w + QString::fromLatin1(" FOR UPDATE"));
    qry.setForwardOnly(true);
    if (!qry.exec())
      return QVariantList();

    QVariantList ret;
    int countFields = qry.fieldList().count();

    ret.append(qry.size());
    ret.append(countFields);
    ret.append(qry.fieldList());

    while (qry.next()) {
      for (int i = 0; i < countFields; ++i)
        ret.append(qry.value(i));
    }

    return ret;
  }
};
"""



"""QLine module."""
# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore[import]
from typing import Optional


class QLine(QtWidgets.QFrame):
    """QLine class."""

    _object_name: Optional[str]
    _orientation: int

    def __init__(self, parent) -> None:
        """Inicialize."""

        super().__init__()
        self._object_name = None
        self._orientation = 0

    def getObjectName(self) -> Optional[str]:
        """Return object name."""

        return self._object_name

    def setObjectName(self, name: str) -> None:  # type: ignore [override]
        """Set object name."""

        self._object_name = name

    def setOrientation(self, ori_: int = 0) -> None:
        """Set orientation."""

        self._orientation = ori_
        self.setFrameShape(self.Shape.HLine if ori_ == 1 else self.Shape.VLine)

    def getOrientation(self) -> int:
        """Return orientation."""

        return self._orientation

    orientation = property(getOrientation, setOrientation)
    objectName = property(getObjectName, setObjectName)  # type: ignore [assignment]



"""
File module.
"""
import os
from pineboolib import logging
from typing import Optional

from pineboolib.application.utils import path

LOGGER = logging.get_logger(__name__)


class File(object):
    """
    Manage files from a module.
    """

    module: str
    filename: str
    basedir: Optional[str]
    sha: Optional[str]
    name: str
    ext: str

    def __init__(
        self,
        module: str,
        filename: str,
        sha: Optional[str] = None,
        basedir: Optional[str] = None,
        db_name: Optional[str] = None,
    ) -> None:
        """
        Initialize.

        @param module. Identificador del módulo propietario
        @param filename. Nombre del fichero
        @param sha. Código sha1 del contenido del fichero
        @param basedir. Ruta al fichero en cache
        """
        self.module = module
        self.filename = filename
        self.sha = sha
        self.name, self.ext = os.path.splitext(filename)

        self.filekey = (
            "%s/%s/file%s/%s/%s%s" % (db_name, module, self.ext, self.name, sha, self.ext)
            if self.sha
            else filename
        )

        self.basedir = basedir

    def path(self) -> str:
        """
        Return absolute path to file.

        @return Ruta absoluta del fichero
        """
        return (
            path._dir(self.basedir, self.filename)
            if self.basedir
            else path._dir("cache", *(self.filekey.split("/")))
        )



"""Test pending-replationships module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa


class TestRelationships(unittest.TestCase):
    """TestByteArray Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_list(self):
        """Test list"""

        class_area = qsa.orm_("flareas", False)
        class_modulo = qsa.orm_("flmodules", False)

        self.assertFalse(class_area is None)
        self.assertFalse(class_modulo is None)

        obj_area = class_area()
        obj_area.bloqueo = False
        obj_area.idarea = "T"
        obj_area.descripcion = "Area de pruebas de pendingRelationships"
        obj_area.save()

        obj_modulo_1 = class_modulo()
        obj_modulo_1.bloqueo = False
        obj_modulo_1.idmodulo = "M1"
        obj_modulo_1.idarea = obj_area.idarea
        obj_modulo_1.descripcion = "Modulo de pruebas 1 de pendingrelationships"
        obj_modulo_1.version = "0.1"
        obj_modulo_1.save()

        obj_modulo_1 = class_modulo()
        obj_modulo_1.bloqueo = False
        obj_modulo_1.idmodulo = "M2"
        obj_modulo_1.idarea = obj_area.idarea
        obj_modulo_1.descripcion = "Modulo de pruebas 2 de pendingrelationships"
        obj_modulo_1.version = "0.1"
        obj_modulo_1.save()

        self.assertTrue(hasattr(obj_area, "children"))

        self.assertEqual(len(obj_area.children), 2, "no son dos => %s" % obj_area.children)

    def test_children_before_after_commit(self):
        class_area = qsa.orm_("flareas", False)
        class_modulo = qsa.orm_("flmodules", False)

        self.assertFalse(class_area is None)
        self.assertFalse(class_modulo is None)

        obj_area = class_area()
        obj_area.bloqueo = False
        obj_area.idarea = "T1"
        obj_area.descripcion = "Area de pruebas de pendingRelationships"
        obj_area.save()

        self.assertTrue(not obj_area.children)

        obj_modulo_1 = class_modulo()
        obj_modulo_1.bloqueo = False
        obj_modulo_1.idmodulo = "T2M1"
        obj_modulo_1.idarea = obj_area.idarea
        obj_modulo_1.descripcion = "Modulo de pruebas 1 de T1"
        obj_modulo_1.version = "0.1"
        obj_modulo_1.save()

        obj_modulo_1 = class_modulo()
        obj_modulo_1.bloqueo = False
        obj_modulo_1.idmodulo = "T2M2"
        obj_modulo_1.idarea = obj_area.idarea
        obj_modulo_1.descripcion = "Modulo de pruebas 2 de T1"
        obj_modulo_1.version = "0.1"
        obj_modulo_1.save()

        obj_area.save()

        self.assertTrue(obj_modulo_1 in obj_area.children, "hijos: %s" % obj_area.children)

        obj_modulo_1.idmodulo = "T2M1_1"
        with self.assertRaises(Exception):
            obj_area.save(relations=["children"])

        obj_modulo_1.idmodulo = "T2M1_2"
        obj_area.save(relations=["children"])  # Guardamos!

        self.assertTrue(obj_modulo_1.idmodulo == "T2M1_2", "idmodulo es %s" % obj_modulo_1.idmodulo)
        self.assertTrue(len(obj_area._session.dirty) == 0)
        self.assertEqual(
            obj_modulo_1.idmodulo, qsa.util.sqlSelect("flmodules", "idmodulo", "idmodulo='T2M1_2'")
        )

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



"""Test spinbox module."""
import unittest


class TestSpinBox(unittest.TestCase):
    """TestSpinBox class."""

    def test_basic(self) -> None:
        """Test basic."""

        from pineboolib.q3widgets import spinbox

        sb_ = spinbox.SpinBox()
        sb_.maximum = 100
        sb_.minimum = 50
        sb_.value = 75

        self.assertEqual(sb_.maximum, 100)
        self.assertEqual(sb_.minimum, 50)
        self.assertEqual(sb_.value, 75)

        sb_.label = "Prueba"
        self.assertEqual(sb_.label, "Prueba")



"""Flsqlite module."""


from pineboolib.application.utils import path

from pineboolib import logging, application

from pineboolib.fllegacy import flutil

from pineboolib.interfaces import isqldriver

import os


from typing import Optional, Any, List, Dict, TYPE_CHECKING
from sqlalchemy import create_engine, event, util  # type: ignore [import] # noqa: F821, F401


if TYPE_CHECKING:
    from pineboolib.application.metadata import pntablemetadata
    from sqlalchemy.engine import (  # type: ignore [import]
        base,
    )  # type: ignore [import] # noqa: F401, F821 # pragma: no cover
    from sqlalchemy.orm import (  # type: ignore [import] # noqa: F821, F401
        session,
    )  # pragma: no cover

LOGGER = logging.get_logger(__name__)


class FLSQLITE(isqldriver.ISqlDriver):
    """FLSQLITE class."""

    def __init__(self):
        """Inicialize."""
        super().__init__()
        self.version_ = "0.9"
        self.name_ = "FLsqlite"
        self.error_list = []
        self.alias_ = "SQLite3 (SQLITE3)"
        self.db_filename = ""
        self.mobile_ = True
        self.desktop_file = True
        self._null = ""
        self._text_like = ""
        self._text_cascade = ""
        self._parse_porc = False
        self._can_use_preping = False

        util.deprecations.SILENCE_UBER_WARNING = True  # type: ignore [attr-defined]

        self._sqlalchemy_name = "sqlite"

    def getConn(
        self, name: str, host: str, port: int, usern: str, passw_: str, alternative: bool = False
    ) -> "base.Connection":
        """Return connection."""

        conn_ = None
        main_conn = None
        if "main_conn" in application.PROJECT.conn_manager.connections_dict.keys():
            main_conn = application.PROJECT.conn_manager.mainConn()
            conn_driver = main_conn.driver()
            if self.db_filename == conn_driver.db_filename:
                self._engine = conn_driver._engine
                self._connection = conn_driver._connection
                return self._connection

        if conn_ is None:
            if (
                not os.path.exists("%s/sqlite_databases/" % application.PROJECT.tmpdir)
                and not self.db_filename == ":memory:"
            ):
                os.mkdir("%s/sqlite_databases/" % application.PROJECT.tmpdir)

            self.get_common_params()
            self._engine = create_engine(
                self.loadConnectionString(name, host, port, usern, passw_), **self._queqe_params
            )

            event.listen(self._engine, "connect", self.do_connect)
            event.listen(self._engine, "begin", self.do_begin)
            event.listen(self._engine, "savepoint", self.do_savepoint)

            if application.SHOW_CONNECTION_EVENTS:
                self.listen_engine()

            conn_ = self.connection()

            if not os.path.exists("%s" % self.db_filename) and self.db_filename not in [
                ":memory:",
                "temp_db",
            ]:
                LOGGER.warning("La base de datos %s no existe", self.db_filename)

            if conn_ is not None:
                # self.session()
                self._connection = conn_

        return conn_

    def loadConnectionString(self, name: str, host: str, port: int, usern: str, passw_: str) -> str:
        """Set special config."""

        return "%s:///%s" % (self._sqlalchemy_name, self.db_filename)
        # return "%s:///%s?timeout=10&nolock=1" % (self._sqlalchemy_name, self.db_filename)

    def setDBName(self, name: str):
        """Set DB Name."""

        self._dbname = "temp_db" if name == ":memory:" else name

        self.db_filename = path._dir("sqlite_databases/%s.sqlite3" % self._dbname)

        if name == ":memory:":
            if application.PROJECT._splash:
                application.PROJECT._splash.hide()

            if application.VIRTUAL_DB:
                self.db_filename = name

    def setType(self, type_: str, leng: int = 0) -> str:
        """Return type definition."""

        type_ = type_.lower()
        type_array = {
            "int": "INTEGER",
            "uint": "INTEGER",
            "bool": "BOOLEAN",
            "unlock": "BOOLEAN",
            "double": "FLOAT",
            "time": "VARCHAR(20)",
            "date": "VARCHAR(20)",
            "pixmap": "TEXT",
            "stringlist": "TEXT",
            "string": "VARCHAR",
            "bytearray": "CLOB",
            "timestamp": "DATETIME",
            "json": "JSON",
        }

        res_ = type_array[type_] if type_ in type_array.keys() else ""

        if not res_:
            LOGGER.warning("seType: unknown type %s", type_)
            leng = 0

        return "%s(%s)" % (res_, leng) if leng else res_

    # def process_booleans(self, where: str) -> str:
    #    """Process booleans fields."""

    #    return where.replace("'true'", "1").replace("'false'", "0")

    def sqlCreateTable(
        self, tmd: "pntablemetadata.PNTableMetaData", create_index: bool = True
    ) -> Optional[str]:
        """Return a create table query."""

        if tmd.isQuery():
            return self.sqlCreateView(tmd)

        util = flutil.FLUtil()
        primary_key = ""

        unlocks = 0
        sql_fields: List[str] = []
        for field in tmd.fieldList():
            type_ = field.type()

            sql_field = field.name()

            if type_ == "serial":
                sql_field += " INTEGER"
                if not field.isPrimaryKey():
                    sql_field += " PRIMARY KEY"
            else:
                if type_ == "unlock":
                    unlocks += 1

                    if unlocks > 1:
                        LOGGER.debug("FLManager : No se ha podido crear la tabla " + tmd.name())
                        LOGGER.debug(
                            "FLManager : Hay mas de un campo tipo unlock. Solo puede haber uno."
                        )
                        return None

                sql_field += " %s" % self.setType(type_, field.length())

            if field.isPrimaryKey():
                if not primary_key:
                    sql_field += " PRIMARY KEY"
                    primary_key = field.name()
                else:
                    LOGGER.warning(
                        util.translate(
                            "application",
                            "FLManager : Tabla-> %s ." % tmd.name()
                            + "Se ha intentado poner una segunda clave primaria para el campo %s ,pero el campo %s ya es clave primaria."
                            % (primary_key, field.name())
                            + "Sólo puede existir una clave primaria en FLTableMetaData, use FLCompoundKey para crear claves compuestas.",
                        )
                    )
                    raise Exception(
                        "A primary key (%s) has been defined before the field %s.%s -> %s"
                        % (primary_key, tmd.name(), field.name(), sql_fields)
                    )
            else:
                sql_field += " UNIQUE" if field.isUnique() else ""
                sql_field += " NULL" if field.allowNull() else " NOT NULL"

            sql_fields.append(sql_field)

        sql = "CREATE TABLE %s (%s);" % (tmd.name(), ",".join(sql_fields))

        if tmd.primaryKey() and create_index:
            sql += "CREATE INDEX %s_pkey ON %s (%s)" % (tmd.name(), tmd.name(), tmd.primaryKey())

        return sql

    def recordInfo2(self, table_name: str) -> Dict[str, List[Any]]:
        """Return info from a database table."""

        info = {}
        sql = "PRAGMA table_info('%s')" % table_name

        cursor = self.execute_query(sql)

        for col0, field_name, field_type, allow_null, col4, is_pk in list(
            cursor.fetchall() if cursor else []
        ):
            field_allow_null = allow_null == 0 and is_pk == 0  # type: ignore [comparison-overlap]
            field_primary_key = is_pk == 1  # type: ignore [comparison-overlap]
            field_size = (
                int(field_type[field_type.find("(") + 1 : len(field_type) - 1])
                if field_type.find("VARCHAR(") > -1
                else 0
            )

            info[field_name] = [
                field_name,
                self.decodeSqlType(field_type),
                not field_allow_null,
                int(field_size),
                None,  # field_precision
                None,  # default value
                field_primary_key,
            ]

        return info

    def decodeSqlType(self, type_: str) -> str:
        """Return the specific field type."""

        key = type_
        key = "VARCHAR" if key.find("VARCHAR") > -1 else key

        array_types = {
            "VARCHAR": "string",  # Aqui también puede ser time y date
            "FLOAT": "double",
            "INTEGER": "uint",  # serial
            "BOOLEAN": "bool",  # y unlock
            "TEXT": "stringlist",  # Aquí también puede ser pixmap
            "DATETIME": "timestamp",
            "JSON": "json",
        }

        return array_types[key] if key in array_types else str(type_)

    def tables(self, type_name: str = "", table_name: str = "") -> List[str]:
        """Return a tables list specified by type."""

        table_list: List[str] = []
        result_list: List[Any] = []
        if self.is_open():
            where: List[str] = []
            if type_name in ("Tables", ""):
                where.append("type='table'")
            if type_name in ("Views", ""):
                where.append("type='view'")
            if type_name in ("SystemTables", ""):
                table_list.append("sqlite_master")

            if where:
                and_name = " AND name ='%s'" % (table_name) if table_name else ""

                cursor = self.execute_query(
                    "SELECT name FROM sqlite_master WHERE %s%s ORDER BY name ASC"
                    % (" OR ".join(where), and_name)
                )
                result_list = cursor.fetchall() if cursor else []

            table_list += [item[0] for item in result_list]

        return table_list

    def remove_index(self, metadata: "pntablemetadata.PNTableMetaData", query: Any) -> bool:
        """Remove olds index."""

        if not query.exec_("DROP INDEX IF EXISTS %s_pkey" % metadata.name()):
            return False

        return True

    def connection(self) -> "base.Connection":
        """Retrun connection."""

        if not getattr(self, "_connection", None) or self._connection.closed:
            if getattr(self, "_engine", None):
                self._connection = self._engine.connect().execution_options(autocommit=True)
                self._connection.execute("PRAGMA journal_mode=WAL")
                self._connection.execute("PRAGMA synchronous=NORMAL")

                event.listen(self._engine, "close", self.close_emited)
            else:
                raise Exception("Engine is not loaded!")
        return self._connection

    def get_common_params(self) -> None:
        """Load common params."""

        super().get_common_params()

        self._queqe_params["isolation_level"] = None

    def do_connect(self, dbapi_connection, connection_record):
        """Isolation Level fix."""

        dbapi_connection.isolation_level = None



"""Test_fltabledb module."""

from pineboolib.fllegacy import fltabledb
from pineboolib import application
from pineboolib.fllegacy.tests import fixture_path

from typing import Any

import unittest
from pineboolib.loader.main import init_testing, finish_testing


class TestFLTableDB(unittest.TestCase):
    """Test FLTableDB class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_export_to_ods_1(self) -> None:
        """Test export to ods."""

        application.PROJECT.actions["flareas"].openDefaultForm()

        widget = application.PROJECT.actions[  # type: ignore [attr-defined] # noqa F821
            "flareas"
        ]._master_widget

        if widget is None:
            self.assertTrue(widget)
            return

        form = widget.form
        if form is None:
            self.assertTrue(form)
            return
        fltable: Any = form.findChild(fltabledb.FLTableDB, "tableDBRecords")
        self.assertTrue(fltable)
        fltable.exportToOds()

    def test_export_to_ods_2(self) -> None:
        """Test export to ods."""

        application.PROJECT.actions["flmodules"].openDefaultForm()

        widget = application.PROJECT.actions[  # type: ignore [attr-defined] # noqa F821
            "flmodules"
        ]._master_widget

        if widget is None:
            self.assertTrue(widget)
            return

        form = widget.form
        if form is None:
            self.assertTrue(form)
            return

        fltable: Any = form.findChild(fltabledb.FLTableDB, "tableDBRecords")
        self.assertTrue(fltable)
        fltable.exportToOds()

    def test_order_cols(self) -> None:
        """Test order cols."""

        widget = application.PROJECT.actions[  # type: ignore [attr-defined] # noqa F821
            "flareas"
        ]._master_widget

        if widget is None:
            self.assertTrue(widget)
            return

        form = widget.form
        if form is None:
            self.assertTrue(form)
            return

        fltable: Any = form.findChild(fltabledb.FLTableDB, "tableDBRecords")
        fltable.setOrderCols(["descripcion", "idarea", "bloqueo"])
        self.assertEqual(fltable.orderCols(), ["descripcion", "idarea", "bloqueo"])
        fltable.setOrderCols(["idarea"])
        self.assertEqual(fltable.orderCols(), ["idarea", "descripcion", "bloqueo"])

    # def Test_put_x_col(self) -> None:
    #    """Test put first and second col."""

    #    form = application.project.actions[  # type: ignore [attr-defined] # noqa F821
    #        "flareas"
    #    ].mainform_widget

    #    fltable = form.findChild(fltabledb.FLTableDB, "tableDBRecords")
    #    self.assertEqual(fltable.orderCols(), ["idarea", "descripcion", "bloqueo"])
    #    fltable.putFirstCol(1)
    #    self.assertEqual(fltable.orderCols(), ["descripcion", "idarea", "bloqueo"])
    #    fltable.putFirstCol(1)
    #    self.assertEqual(fltable.orderCols(), ["idarea", "descripcion", "bloqueo"])
    #    fltable.putFirstCol(2)
    #    self.assertEqual(fltable.orderCols(), ["bloqueo", "descripcion", "idarea"])
    #    fltable.putFirstCol("idarea")
    #    self.assertEqual(fltable.orderCols(), ["idarea", "descripcion", "bloqueo"])
    #    fltable.putFirstCol("idarea")
    #    self.assertEqual(fltable.orderCols(), ["idarea", "descripcion", "bloqueo"])
    #    fltable.putFirstCol("descripcion")
    #    self.assertEqual(fltable.orderCols(), ["descripcion", "idarea", "bloqueo"])
    #    fltable.putSecondCol(2)
    #    self.assertEqual(fltable.orderCols(), ["descripcion", "bloqueo", "idarea"])
    #    fltable.putSecondCol(0)
    #    self.assertEqual(fltable.orderCols(), ["bloqueo", "descripcion", "idarea"])
    #    fltable.putSecondCol("bloqueo")
    #    self.assertEqual(fltable.orderCols(), ["descripcion", "bloqueo", "idarea"])
    #    fltable.putSecondCol("bloqueo")
    #    self.assertEqual(fltable.orderCols(), ["descripcion", "bloqueo", "idarea"])

    def test_sort_order(self) -> None:
        """Test sort orders."""

        widget = application.PROJECT.actions[  # type: ignore [attr-defined] # noqa F821
            "flareas"
        ]._master_widget

        if widget is None:
            self.assertTrue(widget)
            return

        form = widget.form
        if form is None:
            self.assertTrue(form)
            return

        fltable: Any = form.findChild(fltabledb.FLTableDB, "tableDBRecords")
        cursor = fltable.cursor()
        fltable.setSortOrder(0, 2)
        self.assertEqual(cursor.sort(), "idarea DESC")
        fltable.setSortOrder(False, 1)
        self.assertEqual(cursor.sort(), "bloqueo DESC")
        fltable.setSortOrder(False, 0)
        self.assertEqual(cursor.sort(), "descripcion DESC")
        self.assertFalse(fltable.isSortOrderAscending())
        fltable.setSortOrder(1, 0)
        self.assertEqual(cursor.sort(), "descripcion ASC")
        fltable.setSortOrder(True, 1)
        self.assertEqual(cursor.sort(), "bloqueo ASC")
        self.assertTrue(fltable.isSortOrderAscending())
        # fltable.putSecondCol("bloqueo")
        # self.assertEqual(fltable.orderCols(), ["idarea", "bloqueo", "descripcion"])

    def test_filter_records(self) -> None:
        """Test filterRecords function."""

        widget = application.PROJECT.actions[  # type: ignore [attr-defined] # noqa F821
            "flareas"
        ]._master_widget

        if widget is None:
            self.assertTrue(widget)
            return

        form = widget.form
        if form is None:
            self.assertTrue(form)
            return

        fltable: Any = form.findChild(fltabledb.FLTableDB, "tableDBRecords")
        cursor = fltable.cursor()

        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("idarea", "A")
        cursor.setValueBuffer("descripcion", "AREA A")
        cursor.setValueBuffer("bloqueo", False)
        cursor.commitBuffer()
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("idarea", "A1")
        cursor.setValueBuffer("descripcion", "AREA A1")
        cursor.setValueBuffer("bloqueo", False)
        cursor.commitBuffer()
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("idarea", "B")
        cursor.setValueBuffer("descripcion", "AREA B")
        cursor.setValueBuffer("bloqueo", False)
        cursor.commitBuffer()
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("idarea", "C")
        cursor.setValueBuffer("descripcion", "AREA C")
        cursor.setValueBuffer("bloqueo", False)
        cursor.commitBuffer()

        self.assertEqual(fltable.cursor().size(), 5)
        self.assertEqual(fltable.orderCols(), ["bloqueo", "idarea", "descripcion"])
        fltable.setOrderCols(["idarea", "descripcion", "bloqueo"])
        fltable.filterRecords("X")
        fltable.refresh()  # Forzamos refresh para emular el refresh delayed
        self.assertEqual(fltable.cursor().size(), 0)
        fltable.filterRecords("A")
        fltable.refresh()
        self.assertEqual(fltable.cursor().size(), 2)

    def test_x_edition_flags(self) -> None:
        """Test edition flags."""

        widget = application.PROJECT.actions[  # type: ignore [attr-defined] # noqa F821
            "flareas"
        ]._master_widget

        if widget is None:
            self.assertTrue(widget)
            return

        form = widget.form
        if form is None:
            self.assertTrue(form)
            return

        fltable: Any = form.findChild(fltabledb.FLTableDB, "tableDBRecords")

        self.assertFalse(fltable.readOnly())
        fltable.setReadOnly(True)
        self.assertTrue(fltable.readOnly())
        fltable.setReadOnly(False)
        self.assertFalse(fltable.readOnly())

        self.assertFalse(fltable.editOnly())
        fltable.setEditOnly(True)
        self.assertTrue(fltable.editOnly())
        fltable.setEditOnly(False)
        self.assertFalse(fltable.editOnly())

        self.assertFalse(fltable.insertOnly())
        fltable.setInsertOnly(True)
        self.assertTrue(fltable.insertOnly())
        fltable.setInsertOnly(False)
        self.assertFalse(fltable.insertOnly())

    def test_x_tab_filter(self) -> None:
        """Test tab filter."""

        # from PyQt6 import QtCore
        widget = application.PROJECT.actions[  # type: ignore [attr-defined] # noqa F821
            "flareas"
        ]._master_widget

        if widget is None:
            self.assertTrue(widget)
            return

        form = widget.form
        if form is None:
            self.assertTrue(form)
            return

        fltable: Any = form.findChild(fltabledb.FLTableDB, "tableDBRecords")
        fltable._filter = ""
        fltable.activeTabFilter()
        fltable.tdbFilterClear()
        widget_cb = fltable._tdb_filter.cellWidget(1, 1)
        widget_cb.setCurrentText(fltable.tr("Igual a Valor"))
        widget_le = fltable._tdb_filter.cellWidget(1, 2)
        widget_le.setText("A")
        fltable.activeTabData()
        fltable.refresh()
        cursor1 = fltable.cursor()
        self.assertEqual(cursor1.size(), 1)

        fltable.activeTabFilter()
        fltable.tdbFilterClear()
        widget_cb_2 = fltable._tdb_filter.cellWidget(1, 1)
        widget_cb_2.setCurrentText(fltable.tr("Contiene Valor"))
        widget_le_2 = fltable._tdb_filter.cellWidget(1, 2)
        widget_le_2.setText("A")
        fltable.activeTabData()
        fltable.refresh()
        cursor2 = fltable.cursor()
        self.assertEqual(cursor2.size(), 2)

        fltable.activeTabFilter()
        fltable.tdbFilterClear()
        widget_cb_3 = fltable._tdb_filter.cellWidget(1, 1)
        widget_cb_3.setCurrentText(fltable.tr("Distinto de Valor"))
        widget_le_3 = fltable._tdb_filter.cellWidget(1, 2)
        widget_le_3.setText("W")
        fltable.activeTabData()
        fltable.refresh()
        cursor3 = fltable.cursor()
        cursor3.first()
        self.assertEqual(cursor3.size(), 5)

        fltable.activeTabFilter()
        fltable.tdbFilterClear()
        widget_cb_4 = fltable._tdb_filter.cellWidget(0, 1)
        widget_cb_4.setCurrentText(fltable.tr("Contiene Valor"))
        widget_chb_1 = fltable._tdb_filter.cellWidget(0, 2)
        widget_chb_1.setChecked(True)
        fltable.activeTabData()
        fltable.refresh()
        cursor4 = fltable.cursor()
        self.assertEqual(cursor4.size(), 1)

        # _label = fltable.cursor().model.headerData(0, QtCore.Qt.Horizontal, QtCore.Qt.DisplayRole)
        fltable.activeTabFilter()
        fltable.tdbFilterClear()
        widget_cb_5 = fltable._tdb_filter.cellWidget(0, 1)
        widget_cb_5.setCurrentText(fltable.tr("Contiene Valor"))
        widget_chb_2 = fltable._tdb_filter.cellWidget(0, 2)
        widget_chb_2.setChecked(False)
        fltable.activeTabData()
        fltable.refresh()
        cursor5 = fltable.cursor()
        self.assertEqual(cursor5.size(), 4)

    def test_cursorRelation(self):
        """Test FLTableDB cursor with cursorRelation."""
        from pineboolib.fllegacy import systype
        import os

        qsa_sys = systype.SysType()
        path = fixture_path("principal.eneboopkg")
        self.assertTrue(os.path.exists(path))
        self.assertTrue(qsa_sys.loadModules(path, False))
        application.PROJECT.actions["flareas"].load_master_widget()

        widget = application.PROJECT.actions[  # type: ignore [attr-defined] # noqa F821
            "flareas"
        ]._master_widget

        if widget is None:
            self.assertTrue(widget)
            return

        form = widget.form
        if form is None:
            self.assertTrue(form)
            return

        table_: Any = fltabledb.FLTableDB(form, "new_fltable")
        table_.cursor_ = form.cursor()
        table_.setTableName("flmodules")
        cursor = form.cursor()
        cursor.select("idarea = 'F'")
        # cursor.bufferChanged.emit("idarea")
        table_.initCursor()
        self.assertEqual(cursor.size(), 1)
        cursor.select("idarea = 'M'")
        self.assertEqual(cursor.size(), 0)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Test for flmodules package."""

import os


def fixture_path(*path: str) -> str:
    """
    Get fixture path for this test folder.
    """
    basedir = os.path.realpath(os.path.dirname(__file__))
    filepath = os.path.join(basedir, "fixtures", *path)
    return filepath



"""Test_utils module."""

import unittest
from pineboolib.loader.main import init_testing
from pineboolib.application.database import utils
from pineboolib.application.database import pnsqlcursor


class TestUtils(unittest.TestCase):
    """TestUtils Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_next_counter(self) -> None:
        """Test next counter."""

        cur_2 = pnsqlcursor.PNSqlCursor("fltest3")
        cur_2.setModeAccess(cur_2.Insert)
        cur_2.refreshBuffer()
        cur_2.setValueBuffer("counter_field", "PR0001")

        self.assertTrue(cur_2.commitBuffer())
        val_1 = utils.next_counter("counter_field", cur_2)
        self.assertEqual(val_1, "PR0002")

    def test_full(self) -> None:
        """Test all options."""

        cur_1 = pnsqlcursor.PNSqlCursor("fltest3")

        val_1 = utils.next_counter("counter_field", cur_1)

        utils.sql_insert(
            "fltest3", "counter_field,string_field", "%s,%s" % (val_1, "Campo de prueba 1")
        )

        self.assertEqual(val_1, "000001")

        val_2 = utils.next_counter("b", "counter_field", cur_1)
        self.assertEqual(val_2, "00001")

        val_3 = utils.next_counter("counter_field", cur_1)
        self.assertEqual(val_3, "000002")

        val_4 = utils.next_counter("ABCD", "counter_field", cur_1)
        self.assertEqual(val_4, "01")

        utils.sql_insert(
            "fltest3", "counter_field,string_field", "%s,%s" % (val_3, "Campo de prueba 2")
        )
        utils.sql_insert(
            "fltest3", "counter_field,string_field", "ABCD%s,%s" % (val_4, "Campo de prueba 4")
        )

        val_5 = utils.next_counter("ABCD", "counter_field", cur_1)
        self.assertEqual(val_5, "02")

        self.assertEqual(
            utils.sql_select("fltest3", "string_field", "counter_field = '%s'" % val_1),
            "Campo de prueba 1",
        )
        self.assertEqual(
            utils.sql_select(
                "fltest3", "string_field", "counter_field = '%s'" % val_3, "fltest3", 0, "default"
            ),
            "Campo de prueba 2",
        )

        self.assertNotEqual(
            utils.sql_select(
                "fltest3", "timezone_field", "counter_field = '%s'" % val_3, "fltest3", 0, "default"
            ),
            "",
        )

        self.assertEqual(
            utils.sql_select(
                "fltest3", "string_field", "counter_field = '%s'" % val_3, ["fltest3"], 0, "default"
            ),
            "Campo de prueba 2",
        )

        self.assertEqual(
            utils.quick_sql_select(
                "fltest3", "string_field", "counter_field = '%s'" % val_3, "default"
            ),
            "Campo de prueba 2",
        )

        self.assertTrue(
            utils.sql_update(
                "fltest3",
                ["string_field"],
                ["Campo de prueba 2 Modificado"],
                "counter_field = '%s'" % val_3,
                "default",
            )
        )

        self.assertEqual(
            utils.quick_sql_select(
                "fltest3", "string_field", "counter_field = '%s'" % val_3, "default"
            ),
            "Campo de prueba 2 Modificado",
        )
        self.assertTrue(utils.quick_sql_delete("fltest3", "counter_field ='%s'" % val_1, "default"))
        self.assertTrue(utils.sql_delete("fltest3", "1=1", "dbAux"))



"""Test_flutil module."""

import unittest

from pineboolib.fllegacy import flsmtpclient
from pineboolib import application


class TestSmtp(unittest.TestCase):
    """TestSmtp class."""

    cli: flsmtpclient.FLSmtpClient

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        cls.cli = flsmtpclient.FLSmtpClient()

    def test_smtp(self) -> None:
        """Test smtp functions."""
        self.cli.status.connect(self.cambio)
        mail_ = "aaa@aaa.aa"
        password_ = "bbbbbbb"

        self.cli.setUser(mail_)
        self.cli.setPassword(password_)

        self.cli.setMailServer("smtp.gmail.com")
        self.cli.setPort(587)
        self.cli.setAuthMethod(self.cli.SendAuthLogin)
        self.cli.setConnectionType(self.cli.TlsConnection)

        self.cli.setFrom(mail_)
        self.cli.setTo(mail_)
        self.cli.setSubject("prueba de correo")
        self.cli.setBody("Esto es una prueba de Pineboo %s" % application.PROJECT.load_version())

        res: bool = self.cli.startSend()
        self.assertFalse(res)  # assertTrue when login is ok

    def cambio(self):
        print(self.cli.lastStatusMsg())



"""Dateedit module."""

from PyQt6 import QtWidgets  # type: ignore[import]

from pineboolib.q3widgets import qdateedit
from pineboolib.application.qsatypes import date as datelib
from typing import Optional


class DateEdit(QtWidgets.QWidget):
    """DateEdit class."""

    label_control: QtWidgets.QLabel
    date_control: qdateedit.QDateEdit

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Initialize."""

        super().__init__(parent)
        layout_ = QtWidgets.QHBoxLayout()
        self.label_control = QtWidgets.QLabel()
        self.date_control = qdateedit.QDateEdit()
        layout_.addWidget(self.label_control)
        layout_.addWidget(self.date_control)
        self.setLayout(layout_)

    def getDate(self) -> datelib.Date:
        """Return Date."""

        return datelib.Date(self.date_control.getDate())

    def setDate(self, date_: datelib.Date) -> None:
        """Set Date."""

        self.date_control.setDate(date_)

    def getLabel(self) -> str:
        """Return Label text."""

        return self.label_control.text()

    def setLabel(self, text_: str) -> None:
        """Set Label text."""

        self.label_control.setText(text_)

    date: datelib.Date = property(getDate, setDate)  # type: ignore [assignment] # noqa: F821
    label: str = property(getLabel, setLabel)  # type: ignore [assignment] # noqa: F821
